/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "fsarchiver.h"
#include "arblkqueue.h"
#include "common.h"

int arblkqueue_init(struct s_arblkqueue *q)
{
	if (!q)
	{	errprintf("q is NULL\n");
		return -1;
	}
	q->head=NULL;
	q->curitemnum=1;
	pthread_mutex_init(&q->mutex, NULL);
	return 0;
}

int arblkqueue_add_block_wait(struct s_arblkqueue *q, char *blkdata, u64 blksize, u64 offset, int status)
{
	while ((arblkqueue_count(q)>QUEUE_SIZE) && (get_terminate()==false))
		usleep(QUEUE_WAIT);

	if (get_terminate()==true)
		return -1;
	else
		return arblkqueue_add_block(q, blkdata, blksize, offset, status);
}

int arblkqueue_add_block(struct s_arblkqueue *q, char *blkdata, u64 blksize, u64 offset, int status)
{
	struct s_arblkitem *item;
	struct s_arblkitem *cur;

	if (!q)
	{	dbgprintf(2, "q is NULL\n");
		return -1;
	}
	
	// create the new item in memory
	item=malloc(sizeof(struct s_arblkitem));
	if (!item)
	{	dbgprintf(2, "out of memory 1\n");
		return -1;
	}
	item->blkdata=blkdata;
	item->type=ARBLKITEM_TYPE_BLOCK;
	item->blksize=blksize;
	item->blkoffset=offset;
	item->status=status;
	item->itemnum=q->curitemnum++;
	item->next=NULL;
	
	pthread_mutex_lock(&q->mutex);
	
	if (q->head==NULL) // if list empty
	{
		q->head=item;
	}
	else // list not empty
	{	for (cur=q->head; (cur!=NULL) && (cur->next!=NULL); cur=cur->next);
		cur->next=item;
	}
	
	pthread_mutex_unlock(&q->mutex);
	
	return 0;
}

// function called by the compression thread when a block has been compressed
int arblkqueue_replace_block(struct s_arblkqueue *q, s64 itemnum, char *newdata, u64 newsize, int newstatus)
{
	struct s_arblkitem *cur;
	
	if (q==NULL)
	{	return -1;
	}

	pthread_mutex_lock(&q->mutex);
	
	if (q->head==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		return -1; // item not found
	}
	
	// test the first item
	cur=q->head;
	if (cur && cur->itemnum==itemnum) // block found
		goto arblkqueue_replace_block_found;
	
	// test the other items
	cur=q->head->next;
	while (cur!=NULL)
	{
		if (cur->itemnum==itemnum) // block found
			goto arblkqueue_replace_block_found;
		else
			cur=cur->next;
	}

	pthread_mutex_unlock(&q->mutex);
	return -1; // not found

arblkqueue_replace_block_found:
	cur->blkdata=newdata;
	cur->blksize=newsize;
	cur->status=newstatus;
	pthread_mutex_unlock(&q->mutex);
	return 0;
}

// the compression thread requires the first block which has not yet been compressed
s64 arblkqueue_get_first_block_todo(struct s_arblkqueue *q, char **blkdata, u64 *size, u64 *offset)
{
	struct s_arblkitem *cur;

	*blkdata=NULL;
	
	if (q==NULL)
		return -1; // error

	pthread_mutex_lock(&q->mutex);
	
	if (q->head==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		return 0; // ok, but nothing found
	}
	
	// test the first item
	for (cur=q->head; cur!=NULL; cur=cur->next)
	{
		if ((cur->type==ARBLKITEM_TYPE_BLOCK) && (cur->status==ARBLKITEM_STATUS_TODO))
		{
			*size=cur->blksize;
			*blkdata=cur->blkdata;
			*offset=cur->blkoffset;
			cur->status=ARBLKITEM_STATUS_PROGRESS;
			pthread_mutex_unlock(&q->mutex);
			return cur->itemnum; // ok, return blk number
		}
	}

	pthread_mutex_unlock(&q->mutex);
	return 0; // ok, not found
}

// the writer thread requires the first block of the queue if it ready to go
s64 arblkqueue_dequeue_first_if_done(struct s_arblkqueue *q, int *type, char **blkdata, u64 *size, u64 *offset, char *magic, struct s_dico **dico)
{
	struct s_arblkitem *cur;
	s64 itemnum;
	
	if (q==NULL)
	{	return 0; // ok but not found
	}

	pthread_mutex_lock(&q->mutex);
	if (q->head==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		return 0; // ok but not found
	}
	
	// test the first item
	cur=q->head;
	if ((cur!=NULL) && (cur->status==ARBLKITEM_STATUS_DONE))
	{
		if (cur->type==ARBLKITEM_TYPE_BLOCK) // item to dequeue is a block
		{
			*size=cur->blksize;
			*blkdata=cur->blkdata;
			*offset=cur->blkoffset;
			*type=cur->type;
			itemnum=cur->itemnum;
			q->head=cur->next;
			pthread_mutex_unlock(&q->mutex);
			return itemnum;
		}
		else if (cur->type==ARBLKITEM_TYPE_DICO) // item to dequeue is a dico
		{
			*dico=cur->dico;
			*size=FDARCHIVER_SIZEOF_MAGIC; // header size
			memcpy(magic, cur->magic, FDARCHIVER_SIZEOF_MAGIC); // header contents
			*type=cur->type;
			itemnum=cur->itemnum;
			q->head=cur->next;
			pthread_mutex_unlock(&q->mutex);
			return itemnum;
		}
		else
		{
			errprintf("invalid item type in queue\n");
		}
	}
	
	pthread_mutex_unlock(&q->mutex);
	return 0;  // ok but not found
}

int arblkqueue_count(struct s_arblkqueue *l)
{
	struct s_arblkitem *cur;
	int count=0;
	
	if (l==NULL)
		return -1;
			
	pthread_mutex_lock(&l->mutex);
	cur=l->head;
	while (cur!=NULL)
	{
		count++;
		cur=cur->next;
	}
	pthread_mutex_unlock(&l->mutex);

	return count;
}

int arblkqueue_count_status(struct s_arblkqueue *l, int status)
{
	struct s_arblkitem *cur;
	int count=0;
	
	if (l==NULL)
		return -1;
			
	pthread_mutex_lock(&l->mutex);
	cur=l->head;
	while (cur!=NULL)
	{
		if (status==ARBLKITEM_STATUS_NULL || cur->status==status)
			count++;
		cur=cur->next;
	}
	pthread_mutex_unlock(&l->mutex);

	return count;
}

int arblkqueue_print(struct s_arblkqueue *l)
{
	struct s_arblkitem *cur;
	int count=0;

	if (l==NULL)
		return -1;

	pthread_mutex_lock(&l->mutex);
	printf("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n");
	cur=l->head;
	while (cur!=NULL)
	{
		printf("item[%.2d] has type=%d\n", count++, cur->type);
		cur=cur->next;
	}
	printf("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n");
	pthread_mutex_unlock(&l->mutex);

	return count;
}

int arblkqueue_destroy(struct s_arblkqueue *l)
{
	struct s_arblkitem *cur;
	struct s_arblkitem *next;

	if (l==NULL)
		return -1;

	pthread_mutex_destroy(&l->mutex);
	cur=l->head;
	while (cur!=NULL)
	{
		next=cur->next;
		free(cur);
		cur=next;
	}
	l->head=NULL;
	return 0;
}

int arblkqueue_add_dico_wait(struct s_arblkqueue *q, struct s_dico *d, char *magic)
{
	while ((arblkqueue_count(q)>QUEUE_SIZE) && (get_terminate()==false))
		usleep(QUEUE_WAIT);

	if (get_terminate()==true)
		return -1;
	else
		return arblkqueue_add_dico(q, d, magic);
}

int arblkqueue_add_dico(struct s_arblkqueue *q, struct s_dico *d, char *magic)
{
	struct s_arblkitem *item;
	struct s_arblkitem *cur;
	
	if (!q)
	{	dbgprintf(2, "q is NULL\n");
		return -1;
	}
	
	// create the new item in memory
	item=malloc(sizeof(struct s_arblkitem));
	if (!item)
	{	dbgprintf(2, "out of memory 1\n");
		return -1;
	}

	snprintf(item->magic, sizeof(item->magic), "%s", magic);
	item->type=ARBLKITEM_TYPE_DICO;
	item->dico=d;
	item->status=ARBLKITEM_STATUS_DONE;
	item->itemnum=q->curitemnum++;
	item->next=NULL;
	
	pthread_mutex_lock(&q->mutex);
	
	if (q->head==NULL) // if list empty
	{
		q->head=item;
	}
	else // list not empty
	{	for (cur=q->head; (cur!=NULL) && (cur->next!=NULL); cur=cur->next);
		cur->next=item;
	}

	pthread_mutex_unlock(&q->mutex);
	
	return 0;
}

struct s_dico *arblkqueue_dequeue_header_wait(struct s_arblkqueue *q, char *expectedmagic)
{
	while ((arblkqueue_count(q)<1) && (get_terminate()==false))
		usleep(QUEUE_WAIT);

	if (get_terminate()==true)
		return NULL;
	else
		return arblkqueue_dequeue_header(q, expectedmagic);
}

// the extract function wants to read headers from the queue
struct s_dico *arblkqueue_dequeue_header(struct s_arblkqueue *q, char *expectedmagic)
{	
	struct s_arblkitem *cur;
	struct s_dico *d;
	u64 itemnum;
	
	if (q==NULL)
	{	return NULL; // ok but not found
	}

	pthread_mutex_lock(&q->mutex);
	if ((cur=q->head)==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		errprintf("dequeue - cannot dequeue since q->head is NULL\n");
		return NULL; // not found
	}
	
	// test the first item
	if (cur->type==ARBLKITEM_TYPE_DICO)
	{
		if (memcmp(cur->magic, expectedmagic, FDARCHIVER_SIZEOF_MAGIC)==0)
		{
			q->head=cur->next;
			d=cur->dico;
			itemnum=cur->itemnum;
			free(cur);
			pthread_mutex_unlock(&q->mutex);
			return d;
		}
		else
		{
			errprintf("wrong magic in that item: expected [%s]\n", expectedmagic);
			pthread_mutex_unlock(&q->mutex);
			return NULL;  // ok but not found
		}
	}
	else
	{
		errprintf("dequeue - wrong type of data in the queue: wanted a dico\n");
		pthread_mutex_unlock(&q->mutex);
		return NULL;  // ok but not found
	}
}

// say whether or not the first item is a block and is ready
int arblkqueue_is_first_block_ready(struct s_arblkqueue *q)
{
	struct s_arblkitem *cur;
	int result;
	
	if (q==NULL)
		return false; // not found
	
	pthread_mutex_lock(&q->mutex);
	result = ((cur=q->head)!=NULL) && (cur->type==ARBLKITEM_TYPE_BLOCK) && (cur->status==ARBLKITEM_STATUS_DONE);
	pthread_mutex_unlock(&q->mutex);
	return result;
}

// the extract function wants to read headers from the queue
int arblkqueue_dequeue_block_wait(struct s_arblkqueue *q, char **blkdata, u64 *blksize, u64 *blkoffset)
{	
	while ((arblkqueue_is_first_block_ready(q)!=true) && (get_terminate()==false))
		usleep(QUEUE_WAIT); //, fprintf(stderr, " + ");
	
	if (get_terminate()==true)
		return -1;
	else
		return arblkqueue_dequeue_block(q, blkdata, blksize, blkoffset);
}

// the extract function wants to read headers from the queue
int arblkqueue_dequeue_block(struct s_arblkqueue *q, char **blkdata, u64 *blksize, u64 *blkoffset)
{
	struct s_arblkitem *cur;
	
	if (q==NULL)
	{	return -1; // not found
	}

	pthread_mutex_lock(&q->mutex);
	if ((cur=q->head)==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		errprintf("dequeue - cannot dequeue since q->head is NULL\n");
		return -1; // not found
	}
	
	// test the first item
	if ((cur->type==ARBLKITEM_TYPE_BLOCK) && (cur->status==ARBLKITEM_STATUS_DONE))
	{
		*blkdata=cur->blkdata;
		*blksize=cur->blksize;
		*blkoffset=cur->blkoffset;
		q->head=cur->next;
		free(cur);
		pthread_mutex_unlock(&q->mutex);
		return 0;
	}
	else
	{
		errprintf("dequeue - wrong type of data in the queue: wanted a block, found dico with magic=[%4s]\n", cur->magic);
		pthread_mutex_unlock(&q->mutex);
		return -1;  // ok but not found
	}
}

// returns true if the first item in queue is a dico or a block which is ready
int arblkqueue_is_first_item_ready(struct s_arblkqueue *q)
{
	struct s_arblkitem *cur;
	int ret=false;
	
	if (q==NULL)
		return false; // not found
	
	pthread_mutex_lock(&q->mutex);
	
	if ((cur=q->head)==NULL)
		ret=false; // list empty
	else if (cur->type==ARBLKITEM_TYPE_DICO)
		ret=true; // a dico is always ready
	else if ((cur->type==ARBLKITEM_TYPE_BLOCK) && (cur->status==ARBLKITEM_STATUS_DONE))
		ret=true; // a block which has been prepared is ready
	else // other cases: block not yet prepared
		ret=false;
	
	pthread_mutex_unlock(&q->mutex);
	return ret;
}

// say what the next item which is ready in the queue is but do not remove it
int arblkqueue_check_next_item_wait(struct s_arblkqueue *q, int *type, char *magic)
{
	memset(magic, 0, FDARCHIVER_SIZEOF_MAGIC+1);

	while ((arblkqueue_is_first_item_ready(q)==false) && (get_terminate()==false))
		usleep(QUEUE_WAIT);
		
	if (get_terminate()==true)
		return -1;
	else
		return arblkqueue_check_next_item(q, type, magic);
}

// say what the next item which is ready in the queue is but do not remove it
int arblkqueue_check_next_item(struct s_arblkqueue *q, int *type, char *magic)
{
	struct s_arblkitem *cur;
	
	if (q==NULL)
	{	return -1; // ok but not found
	}

	pthread_mutex_lock(&q->mutex);
	if (q->head==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		return -1; // ok but not found
	}
	
	// test the first item
	cur=q->head;
	if ((cur!=NULL) && (cur->status==ARBLKITEM_STATUS_DONE))
	{
		if (cur->type==ARBLKITEM_TYPE_BLOCK) // item to dequeue is a block
		{
			*type=cur->type;
			memset(magic, 0, FDARCHIVER_SIZEOF_MAGIC);
			pthread_mutex_unlock(&q->mutex);
			return 0;
		}
		else if (cur->type==ARBLKITEM_TYPE_DICO) // item to dequeue is a dico
		{
			memcpy(magic, cur->magic, FDARCHIVER_SIZEOF_MAGIC); // header contents
			*type=cur->type;
			pthread_mutex_unlock(&q->mutex);
			return 0;
		}
		else
		{
			errprintf("invalid item type in queue\n");
			return -1;
		}
	}
	
	pthread_mutex_unlock(&q->mutex);
	return -1;  // not found
}

// destroy the first item in the queue (similar to dequeue but do not read it)
int arblkqueue_destroy_first_item(struct s_arblkqueue *q)
{
	struct s_arblkitem *cur;
	
	if (q==NULL)
	{	return -1; // not found
	}

	pthread_mutex_lock(&q->mutex);
	if ((cur=q->head)==NULL)
	{	pthread_mutex_unlock(&q->mutex);
		errprintf("dequeue - cannot dequeue since q->head is NULL\n");
		return -1; // not found
	}
	
	q->head=cur->next;
	free(cur);
	pthread_mutex_unlock(&q->mutex);
	return 0;
}
