/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <endian.h>
#include <byteswap.h>
#include <linux/limits.h>
#include <sys/mount.h>
#include <sys/statvfs.h>
#include <attr/xattr.h>
#include <zlib.h>
#include <bzlib.h>

#include "types.h"
#include "dico.h"
#include "fsarchiver.h"
#include "archive.h"
#include "common.h"
#include "create.h"
#include "md5.h"
#include "mount.h"
#include "fs_ext2.h"
#include "fs_xfs.h"
#include "fs_reiserfs.h"
#include "fs_reiser4.h"
#include "fs_jfs.h"
#include "fs_btrfs.h"
#include "thread_comp.h"
#include "thread_imgdisk.h"

int backup_obj_regfile(struct s_archinfo *ai, char *relpath, char *fullpath, u64 filesize)
{
	struct md5_ctx md5ctx;
	struct s_dico *footerdico;
	u64 curblocksize;
	u8 *origblock;
	u8 md5sum[16];
	u64 filepos;
	int ret=0;
	int res;
	int fd;
	
	// init
	if ((footerdico=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}
	
	md5_init_ctx(&md5ctx);
	
	fd=open(fullpath, O_RDONLY);
	if (fd<0)
	{	sysprintf("Cannot open %s for reading\n", relpath);
		return -1;
	}
	
	for (filepos=0; filepos<filesize; filepos+=curblocksize)
	{
		curblocksize=min(filesize-filepos, DATA_DEF_BLOCK_SIZE);
		
		origblock=malloc(DATA_MAX_BLOCK_SIZE);
		if (!origblock)
		{	errprintf ("malloc() failed: cannot allocate data block\n");
			return -1;
		}

		if (read(fd, origblock, (long)curblocksize)!=curblocksize)
		{	sysprintf("Cannot read data block from %s, block=%ld and res=%ld\n", relpath, (long)curblocksize, (long)res);
			ret=-1;
			goto backup_regfile_error;
		}
		
		md5_process_bytes(origblock, curblocksize, &md5ctx);
		
		if (arblkqueue_add_block_wait(&ai->queue, (char*)origblock, curblocksize, filepos, ARBLKITEM_STATUS_TODO)!=0)
		{	sysprintf("arblkqueue_add_block_wait(%s) failed\n", relpath);
			ret=-1;
			goto backup_regfile_error;
		}
	}
	
	// write the footer with the global md5sum
	md5_finish_ctx(&md5ctx, md5sum);
	dico_add(footerdico, BLOCKFOOTITEMKEY_MD5SUM, md5sum, 16);
	if (arblkqueue_add_dico_wait(&ai->queue, footerdico, FDARCHIVER_MAGIC_FILEFOOTER)!=0)
	{	dbgprintf(1, "Cannot write footer for file %s\n", relpath);
		ret=-1;
		goto backup_regfile_error;
	}
	
backup_regfile_error:
	close(fd);
	return ret;
}

int backup_item_xattr(struct s_archinfo *ai, char *root, char *relpath, struct stat *statbuf, struct s_dico *d)
{
	char fullpath[PATH_MAX];
	char buffer[4096];
	char value[1024];
	size_t valsize;
	int listlen;
	u64 attrcnt;
	int pos;
	int len;

	// init
	concatenate_paths(fullpath, sizeof(fullpath), root, relpath);
	attrcnt=0;

	memset(buffer, 0, sizeof(buffer));
	listlen=llistxattr(fullpath, buffer, sizeof(buffer)-1);
	
	for (pos=0; (pos<listlen) && (pos<sizeof(buffer)); pos+=len)
	{
		len=strlen(buffer+pos)+1;
		memset(value, 0, sizeof(value));
		valsize=lgetxattr(fullpath, buffer+pos, value, sizeof(value));
		msgprintf(2, "            lgetxattr(%s)=%d: [%s]=[%s]\n", relpath, valsize, buffer+pos, value);
		dico_add_string(d, (2*attrcnt)+0, buffer+pos);
		dico_add(d, (2*attrcnt)+1, value, valsize);
		attrcnt++;
	}
	
	return 0;
}

int backup_item_stdattr(struct s_archinfo *ai, char *root, char *relpath, struct stat *statbuf, struct s_dico *d, int *objtype, struct s_dico *dichardlinks)
{
	char fullpath[PATH_MAX];
	char buffer[PATH_MAX];
	int res;

	// init
	*objtype=OBJTYPE_NULL;
	concatenate_paths(fullpath, sizeof(fullpath), root, relpath);

	// 1. dico += common properties (name, permissions, ...)
	if (dico_add_string(d, DISKITEMKEY_PATH, relpath)!=0)
	{	dbgprintf(2, "dico_add_string(path) failed\n");
		return -1;
	}
	if (dico_add_u64(d, DISKITEMKEY_SIZE, (u64)statbuf->st_size)!=0)
	{	dbgprintf(2,"dico_add_u64(filesize) failed\n");
		return -1;
	}
	if (dico_add_u32(d, DISKITEMKEY_MODE, (u32)statbuf->st_mode)!=0)
	{	dbgprintf(2,"dico_add_u32(mode) failed\n");
		return -1;
	}
	if (dico_add_u32(d, DISKITEMKEY_UID, (u32)statbuf->st_uid)!=0)
	{	dbgprintf(2,"dico_add_u32(uid) failed\n");
		return -1;
	}
	if (dico_add_u32(d, DISKITEMKEY_GID, (u32)statbuf->st_gid)!=0)
	{	dbgprintf(2,"dico_add_u32(gid) failed\n");
		return -1;
	}
	if (dico_add_u64(d, DISKITEMKEY_ATIME, (u32)statbuf->st_atime)!=0)
	{	dbgprintf(2,"dico_add_u32(atime) failed\n");
		return -1;
	}
	if (dico_add_u64(d, DISKITEMKEY_MTIME, (u32)statbuf->st_mtime)!=0)
	{	dbgprintf(2,"dico_add_u32(mtime) failed\n");
		return -1;
	}

	// 2. copy specific properties to the dico
	switch (statbuf->st_mode & S_IFMT)
	{
		case S_IFDIR:
			*objtype=OBJTYPE_DIR;
			break;
		case S_IFLNK:
			*objtype=OBJTYPE_SYMLINK;
			memset(buffer, 0, sizeof(buffer));
			if ((readlink(fullpath, buffer, sizeof(buffer)))<0)
				return -1;
			dico_add_string(d, DISKITEMKEY_SYMLINK, buffer);
			break;
		case S_IFREG:
			if (statbuf->st_nlink==1)
				*objtype=OBJTYPE_REGFILE;
			else // there are hard links
			{	
				res=dico_get_string(dichardlinks, (u64)statbuf->st_ino, buffer, sizeof(buffer));
				if (res!=0) // first time we hit this inode --> regular file
				{	dico_add_string(dichardlinks, (u64)statbuf->st_ino, relpath);
					*objtype=OBJTYPE_REGFILE;
				} else // inode already seen --> hard link
				{	dico_add_string(d, DISKITEMKEY_HARDLINK, buffer);
					*objtype=OBJTYPE_HARDLINK;
				}
			}
			break;
		case S_IFCHR:
			dico_add_u64(d, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_CHARDEV;
			break;
		case S_IFBLK:
			dico_add_u64(d, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_BLOCKDEV;
			break;
		case S_IFIFO:
			dico_add_u64(d, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_FIFO;
			break;
		case S_IFSOCK:
			dico_add_u64(d, DISKITEMKEY_RDEV, statbuf->st_rdev);
			*objtype=OBJTYPE_SOCKET;
			break;
		default:
			errprintf("unknown item %s\n", fullpath);
			return -1;
			break;
	}

	dico_add_u32(d, DISKITEMKEY_OBJTYPE, *objtype);

	return 0;
}

int backup_file(struct s_archinfo *ai, char *root, char *relpath, struct stat *statbuf, struct s_dico *dichardlinks)
{
	char fullpath[PATH_MAX];
	struct s_dico *d;
	int objtype;
	int res;

	// ---- backup standard file attributes
	if ((d=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}

	res=backup_item_stdattr(ai, root, relpath, statbuf, d, &objtype, dichardlinks);
	if (res!=0)
	{	errprintf("cannot prepare header for item %s\n", relpath);
		dico_destroy(d);
		return -1;
	}
	arblkqueue_add_dico_wait(&ai->queue, d, FDARCHIVER_MAGIC_OBJT);
	if (get_terminate()==false)
		msgprintf(1, "-[%s] %s\n", get_objtype_name(objtype), relpath);

	// ---- backup extended file attributes
	if ((d=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}

	res=backup_item_xattr(ai, root, relpath, statbuf, d);
	if (res!=0)
	{	errprintf("cannot prepare xattr-dico for item %s\n", relpath);
		dico_destroy(d);
		return -1;
	}

	arblkqueue_add_dico_wait(&ai->queue, d, FDARCHIVER_MAGIC_XATTR);

	// ---- backup file contents for regfiles
	if (objtype==OBJTYPE_REGFILE)
	{
		concatenate_paths(fullpath, sizeof(fullpath), root, relpath);
		if (backup_obj_regfile(ai, relpath, fullpath, statbuf->st_size)!=0)
		{	errprintf("backup_obj_regfile(%s) failed\n", relpath);
			return -1;
		}
	}

	return 0;
}

int backup_dir(struct s_archinfo *ai, char *root, char *path, struct s_dico *dichardlinks, u64 dev)
{	
	char fulldirpath[PATH_MAX];
	char fullpath[PATH_MAX];
	char relpath[PATH_MAX];
	struct stat statbuf;
	struct dirent *dir;
	DIR *dirdesc;

	// init 
	concatenate_paths(fulldirpath, sizeof(fulldirpath), root, path);

	dirdesc = opendir(fulldirpath);
	if (!dirdesc)
	{       sysprintf ("cannot open directory %s\n", fulldirpath);
		return -1;
	}
	
	while (((dir = readdir(dirdesc)) != NULL) && (get_terminate()==0))
	{
		// ---- calculate paths
		concatenate_paths(relpath, sizeof(relpath), path, dir->d_name);
		concatenate_paths(fullpath, sizeof(fullpath), fulldirpath, dir->d_name);
		if (lstat(fullpath, &statbuf)!=0)
		{	sysprintf ("cannot stat %s\n", fullpath);
			closedir(dirdesc);
			return -1;
		}

		// ---- ignore "." and ".." and ignore mount-points
		if (strcmp(dir->d_name,".")==0 || strcmp(dir->d_name,"..")==0 || ((u64)(statbuf.st_dev)!=dev))
			continue; // ignore "." and ".."
		
		// backup contents before the directory itself so that the dir-attributes are written after the dir contents
		if (S_ISDIR(statbuf.st_mode))
		{
			if (backup_dir(ai, root, relpath, dichardlinks, dev)!=0)
			{	errprintf("backup_dir(%s) failed\n", relpath);
				goto backup_dir_err;
			}
		}
		else // not a directory
		{	
			if (backup_file(ai, root, relpath, &statbuf, dichardlinks)!=0)
			{	errprintf("backup_file(%s) failed\n", relpath);
				goto backup_dir_err;
			}
		}
	}
	
	// backup the directory itself (important for the root inode)
	if (lstat(fulldirpath, &statbuf)!=0)
	{	sysprintf ("cannot stat %s\n", fulldirpath);
		goto backup_dir_err;
	}
	if (backup_file(ai, root, path, &statbuf, dichardlinks)!=0)
	{	errprintf("backup_file(%s,%s) failed\n", root, path);
		goto backup_dir_err;
	}

backup_dir_err:
	closedir(dirdesc);
	return 0;
}

int write_mainhead(struct s_archinfo *ai, char *label)
{
	struct timeval now;
	struct s_dico *d;

	if (!ai)
	{	dbgprintf(2, "ai is NULL\n");
		return -1;
	}
	
	// init
	gettimeofday(&now, NULL);
	if ((d=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}

	dico_add_string(d, MAINHEADKEY_FILEFORMATVER, FSARCHIVER_FILEFORMAT);
	dico_add_string(d, MAINHEADKEY_PROGVERCREAT, FSARCHIVER_VERSION);
	dico_add_string(d, MAINHEADKEY_ARCHLABEL, label);
	dico_add_u64(d, MAINHEADKEY_CREATTIME, now.tv_sec);
	dico_add_u32(d, MAINHEADKEY_ARCHIVEID, ai->archid);
	
	if (arblkqueue_add_dico_wait(&ai->queue, d, FDARCHIVER_MAGIC_MAIN)!=0)
	{	errprintf("cannot write dico for main header\n");
		dico_destroy(d);
		return -1;
	}

	return 0;
}

int archive_filesystem(struct s_archinfo *ai, char *partition)
{
	struct s_dico *dichardlinks;
	struct s_dico *dicofshead;
	struct s_dico *dicoend;
	struct stat statbuf;
	struct statvfs statfsbuf;
	u64 fsbytestotal;
	u64 fsbytesused;
	int mountedbyfsa;
	char mntbuf[PATH_MAX];
	char optbuf[128];
	char fsbuf[128];
	int readwrite;
	int res;
	int ret;

	res=mount_get_devinfo(partition, &readwrite, mntbuf, sizeof(mntbuf), optbuf, sizeof(optbuf), fsbuf, sizeof(fsbuf));
	if (res==0) // partition is mounted
	{
		mountedbyfsa=0;
		dbgprintf(1, "mount_get_devinfo(%s): mnt=[%s], opt=[%s], fs=[%s], rw=[%d]\n", partition, mntbuf, optbuf, fsbuf, readwrite);
		if (readwrite==1 && g_options.allowsaverw==0)
		{
			errprintf("partition [%s] is mounted read/write. Please mount it read-only and \n"
				"try again. you can do \"mount -o remount,ro %s\"\n", partition, partition);
			return -1;
		}
	}
	else // partition not mounted
	{
		mountedbyfsa=1;
		dbgprintf(1, "partition %s is not mounted\n", partition);
		generate_random_tmpdir(mntbuf, sizeof(mntbuf));
		mkdir_recursive(mntbuf);
		memset(fsbuf, 0, sizeof(fsbuf));
		res=mount_partition(partition, mntbuf, fsbuf, sizeof(fsbuf), MS_RDONLY);
		if (res!=0)
		{
			errprintf("partition [%s] cannot be mounted on %s. cannot continue.\n", partition, mntbuf);
			return -1;
		}
	}

	// get space statistics
	if (statvfs(mntbuf, &statfsbuf)!=0)
	{	errprintf ("statvfs(%s) failed\n", mntbuf);
		return -1;
	}
	fsbytestotal=(u64)statfsbuf.f_frsize*(u64)statfsbuf.f_blocks;
	fsbytesused=fsbytestotal-((u64)statfsbuf.f_frsize*(u64)statfsbuf.f_bfree);
	
	// add generic details to the fs-header dico
	//dico_init(dicofshead, FDARCHIVER_SIZEOFKEY_FSYSHEAD);
	if ((dicofshead=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}

	dico_add_string(dicofshead, FSYSHEADKEY_FILESYSTEM, fsbuf);
	dico_add_string(dicofshead, FSYSHEADKEY_MNTPATH, mntbuf);
	dico_add_u64(dicofshead, FSYSHEADKEY_BYTESTOTAL, fsbytestotal);
	dico_add_u64(dicofshead, FSYSHEADKEY_BYTESUSED, fsbytesused);
	
	// add fs-specific info: (label, uuid, block size, ...)  to the fs-header dico
	if ((strcmp(fsbuf, "ext2")==0) || (strcmp(fsbuf, "ext3")==0) || (strcmp(fsbuf, "ext4")==0))
	{
		if (ext2_get_specific_features(dicofshead, partition)!=0)
		{	errprintf ("cannot save filesystem attributes for partition %s\n", partition);
			return -1;
		}
	}
	else if (strcmp(fsbuf, "reiserfs")==0)
	{
		if (reiserfs_get_specific_features(dicofshead, partition)!=0)
		{	errprintf ("cannot save filesystem attributes for partition %s\n", partition);
			return -1;
		}
	}
	else if (strcmp(fsbuf, "reiser4")==0)
	{
		if (reiser4_get_specific_features(dicofshead, partition)!=0)
		{	errprintf ("cannot save filesystem attributes for partition %s\n", partition);
			return -1;
		}
	}
	else if (strcmp(fsbuf, "btrfs")==0)
	{
		if (btrfs_get_specific_features(dicofshead, partition)!=0)
		{	errprintf ("cannot save filesystem attributes for partition %s\n", partition);
			return -1;
		}
	}
	else if (strcmp(fsbuf, "xfs")==0)
	{
		if (xfs_get_specific_features(dicofshead, partition)!=0)
		{	errprintf ("cannot save filesystem attributes for partition %s\n", partition);
			return -1;
		}
	}
	else if (strcmp(fsbuf, "jfs")==0)
	{
		if (jfs_get_specific_features(dicofshead, partition)!=0)
		{	errprintf ("cannot save filesystem attributes for partition %s\n", partition);
			return -1;
		}
	}
	
	// write fs-header dico to the archive file
	if (arblkqueue_add_dico_wait(&ai->queue, dicofshead, FDARCHIVER_MAGIC_FSYS)!=0)
	{	errprintf("cannot write dico for main header\n");
		dico_destroy(dicofshead);
		return -1;
	}

	// get dev ids of the filesystem to ignore mount points
	if (lstat(mntbuf, &statbuf)!=0)
	{	sysprintf ("cannot stat %s\n", mntbuf);
		return -1;
	}

	if ((dichardlinks=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}

	ret=backup_dir(ai, mntbuf, "/", dichardlinks, (u64)statbuf.st_dev);

	// dico for hard links not required anymore
	dico_destroy(dichardlinks);

	// write objtype=OBJTYPE_END
	if ((dicoend=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}

	dico_add_u32(dicoend, DISKITEMKEY_OBJTYPE, OBJTYPE_END);
	arblkqueue_add_dico_wait(&ai->queue, dicoend, FDARCHIVER_MAGIC_OBJT);

	// unmount the partition if necessary
	if (mountedbyfsa==1)
	{
		if (unmount_partition(mntbuf)!=0)
			sysprintf("cannot umount %s\n", mntbuf);
	}
	
	return ret;
}

int do_create(char *archive, char **partition)
{
	pthread_t thread_comp[FDARCHIVER_MAXJOBS];
	pthread_t thread_writer;
	struct s_archinfo ai;
	struct stat st;
	int ret=0;
	int i;
	
	// init
	archive_init(&ai);
	
	// check the archive does not exist
	if ((g_options.overwrite==0) && (regfile_exists(archive)==1))
	{	errprintf("%s already exists, please remove it first.\n", archive);
		ret=-1;
		goto do_create_destroy;
	}
	
	if (archive_create(&ai, archive)!=0)
	{       errprintf ("archive_create(%s) failed\n", archive);
		ret=-1;
		goto do_create_destroy;
	}
	
	// create compression threads
	for (i=0; (i<g_options.compressjobs) && (i<FDARCHIVER_MAXJOBS); i++)
	{
		if (pthread_create(&thread_comp[i], NULL, thread_comp_fct, (void*)&ai) != 0)
		{       errprintf ("pthread_create(thread_comp_fct) failed\n");
			ret=-1;
			goto do_create_destroy;
		}
	}
	
	// create archive-writer thread
	if (pthread_create(&thread_writer, NULL, thread_writer_fct, (void*)&ai) != 0)
	{       errprintf ("pthread_create(thread_writer_fct) failed\n");
		ret=-1;
		goto do_create_destroy;
	}
	
	if (write_mainhead(&ai, "archive-label")!=0)
	{       errprintf ("archive_write_mainhead(%s) failed\n", archive);
		ret=-1;
		goto do_create_close;
	}
	
	// check the partitions are all block devices
	for (i=0; partition[i]; i++)
	{
		if ((stat(partition[i], &st)!=0) || (!S_ISBLK(st.st_mode)))
		{	errprintf("%s is not a valid block device\n", partition[i]);
			goto do_create_close;
		}
	}
	
	for (i=0; partition[i]; i++)
	{
		if (archive_filesystem(&ai, partition[i])!=0)
		{	errprintf("archive_filesystem(%s) failed\n", partition[i]);
			goto do_create_close;
		}
	}
	
	ai.terminate=true; // ask threads to terminate
	
	for (i=0; (i<g_options.compressjobs) && (i<FDARCHIVER_MAXJOBS); i++)
	{
		if (pthread_join(thread_comp[i], NULL) != 0)
		{       errprintf ("pthread_join(thread_comp) failed\n");
			ret=-1;
			goto do_create_destroy;
		}
	}
		
	if (pthread_join(thread_writer, NULL) != 0)
	{       errprintf ("pthread_join(thread_writer) failed\n");
		ret=-1;
		goto do_create_destroy;
	}
	
do_create_close:
	archive_close(&ai);
do_create_destroy:
	archive_destroy(&ai);
	return ret;
}
