/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include "dico.h"
#include "fsarchiver.h"
#include "common.h"

struct s_dico *dico_alloc()
{
	struct s_dico *d;
	if ((d=malloc(sizeof(struct s_dico)))==NULL)
		return NULL;
	d->head=NULL;
	return d;
}

int dico_destroy(struct s_dico *d)
{
	struct s_item *item, *next;
	
	if (d==NULL)
	{	dbgprintf(1, "invalid dictionnary\n");
		return -1;
	}
	
	item=d->head;
	d->head=NULL;
	while (item!=NULL)
	{
		next=item->next;
		free(item->data);
		free(item);
		item=next;
	}
	
	return 0;
}

int dico_add(struct s_dico *d, u64 key, const void *data, u16 size)
{
	struct s_item *item, *lnew, *last;

	if (size<1 || size>DICO_MAX_DATASIZE)
		return -1;

	if (d==NULL)
	{	dbgprintf(1, "invalid dictionnary\n");
		return -2;
	}

	// go to the end of the item and check for duplicates
	if (d->head==NULL) // item is empty
	{	
		lnew=malloc(sizeof(struct s_item));
		if (!lnew)
		{	dbgprintf(1, "out of memory\n");
			return -DICO_EMEM;
		}
		memset(lnew, 0, sizeof(struct s_item));
		d->head=lnew;
	}
	else // item is not empty
	{	
		for (item=d->head; item!=NULL; item=item->next)
		{	last=item;
			//if ((item->name) && (strcmp(item->name, name)==0))
			if (item->key==key)
				return -DICO_EDUPLICATE;
		}
		lnew=malloc(sizeof(struct s_item));
		if (!lnew)
		{	dbgprintf(1, "out of memory\n");
			return -DICO_EMEM;
		}
		memset(lnew, 0, sizeof(struct s_item));
		last->next=lnew;
	}

	// copy key
	lnew->key=key;

	// allocate memory for data
	lnew->data=malloc(size);
	if (!lnew->data)
		return -DICO_EMEM; // out of memory

	// copy data
	memcpy(lnew->data, data, size);
	lnew->size=size;
	return 0;
}

int dico_get(struct s_dico *d, u64 key, void *data, u16 maxsize, u16 *size)
{
	struct s_item *item;
	
	if (size!=NULL)
		*size=0;

	if (d==NULL)
	{	dbgprintf(1, "invalid dictionnary\n");
		return -1;
	}
	
	if (d->head==NULL)
	{	dbgprintf(1, "dico is empty\n");
		return -2;
	}
	
	if (maxsize<1 || maxsize>DICO_MAX_DATASIZE)
		return -DICO_EBADSIZE;
	
	for (item=d->head; item!=NULL; item=item->next)
	{	
		if ((item!=NULL) && (item->key==key))
		{
			if (item->size > maxsize) // item is too big
				return -3;
			memcpy(data, item->data, item->size);
			if (size!=NULL)
				*size=item->size;
			return 0;
		}
	}
	
	return -4; // not found
}

int dico_count(struct s_dico *d)
{
	struct s_item *item;
	int count;

	if (d==NULL)
	{	dbgprintf(1, "invalid dictionnary\n");
		return -DICO_EINVAL;
	}
	
	count=0;
	for (item=d->head; item!=NULL; item=item->next)
		count++;

	return count;
}

int dico_add_u32(struct s_dico *d, u64 key, u32 data)
{
	u32 ledata;
	ledata=cpu_to_le32(data);
	return dico_add(d, key, &ledata, sizeof(ledata));
}

int dico_add_s32(struct s_dico *d, u64 key, s32 data)
{
	s32 ledata;
	ledata=cpu_to_le32(data);
	return dico_add(d, key, &ledata, sizeof(ledata));
}

int dico_add_u64(struct s_dico *d, u64 key, u64 data)
{
	u64 ledata;
	ledata=cpu_to_le64(data);
	return dico_add(d, key, &ledata, sizeof(ledata));
}

int dico_add_s64(struct s_dico *d, u64 key, s64 data)
{
	s64 ledata;
	ledata=cpu_to_le64(data);
	return dico_add(d, key, &ledata, sizeof(ledata));
}

int dico_add_string(struct s_dico *d, u64 key, const char *szstring)
{
	u16 len;
	len=strlen(szstring);
	return dico_add(d, key, szstring, len+1);
}

int dico_get_u32(struct s_dico *d, u64 key, u32 *data)
{
	u32 ledata;
	u16 size;

	*data=0;
	if (dico_get(d, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le32_to_cpu(ledata);
	return 0;
}

int dico_get_s32(struct s_dico *d, u64 key, s32 *data)
{
	s32 ledata;
	u16 size;

	*data=0;
	if (dico_get(d, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le32_to_cpu(ledata);
	return 0;
}

int dico_get_u64(struct s_dico *d, u64 key, u64 *data)
{
	u64 ledata;
	u16 size;

	*data=0;
	if (dico_get(d, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le64_to_cpu(ledata);
	return 0;
}

int dico_get_s64(struct s_dico *d, u64 key, s64 *data)
{
	s64 ledata;
	u16 size;

	*data=0;
	if (dico_get(d, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le64_to_cpu(ledata);
	return 0;
}

int dico_get_string(struct s_dico *d, u64 key, char *buffer, u16 bufsize)
{
	u16 size;
	memset(buffer, 0, bufsize);
	return dico_get(d, key, buffer, bufsize, &size);
}

int dico_show(struct s_dico *d)
{
	struct s_item *item;
	char *dat;
	int i;
	
	printf("\n-----------------debug-dico-begin---------------\n");

	if (d->head)
	{
		for (item=d->head; item!=NULL; item=item->next)
		{
			if ((item->data!=NULL) && (item->size>0))
			{
				printf("key=[%ld], sizeof(data)=[%d], data=[", (long)item->key, (int)item->size);
				dat=item->data;
				for (i=0; i<item->size; i++)
					printf("%.2x", dat[i]);
				printf("]=[%s]\n", dat);
			}
		}
	}
	else
	{
		printf("dico is empty\n");
	}

	printf("-----------------debug-dico-end------------------\n\n");
	return 0;
}

