/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <endian.h>
#include <byteswap.h>
#include <linux/limits.h>
#include <utime.h>
#include <attr/xattr.h>
#include <sys/statvfs.h>

#include "types.h"
#include "fsarchiver.h"
#include "dico.h"
#include "common.h"
#include "extract.h"
#include "archive.h"
#include "md5.h"
#include "mount.h"
#include "fs_ext2.h"
#include "fs_reiserfs.h"
#include "fs_reiser4.h"
#include "fs_btrfs.h"
#include "fs_xfs.h"
#include "fs_jfs.h"
#include "thread_comp.h"
#include "thread_imgdisk.h"

int restore_obj_symlink(struct s_archinfo *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	char buffer[PATH_MAX];

	if (dico_get_string(d, DISKITEMKEY_SYMLINK, buffer, PATH_MAX)<0)
	{	errprintf("Cannot read field=symlink for file=[%s]\n", fullpath);
		return -1;
	}

	if (symlink(buffer, fullpath)<0)
	{	sysprintf("symlink(%s, %s) failed\n", buffer, fullpath);
		return -1;
	}
	return 0;
}

int restore_obj_hardlink(struct s_archinfo *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	char buffer[PATH_MAX];
	char regfile[PATH_MAX];

	if (dico_get_string(d, DISKITEMKEY_HARDLINK, buffer, PATH_MAX)<0)
		return -1;
	concatenate_paths(regfile, PATH_MAX, destdir, buffer);
	if (link(regfile, fullpath)!=0)
		return -2;

	return 0;
}

int restore_obj_devfile(struct s_archinfo *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	dev_t dev;
	u64 temp;
	u32 mode;

	if (dico_get_u64(d, DISKITEMKEY_RDEV, &temp)!=0)
		return -1;
	if (dico_get_u32(d, DISKITEMKEY_MODE, &mode)!=0)
		return -2;
	dev=temp;
	if (mknod(fullpath, mode, dev)!=0)
		return -3;
	return 0;
}

int restore_obj_directory(struct s_archinfo *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	mkdir_recursive(fullpath);
	return 0;
}

int restore_obj_regfile(struct s_archinfo *ai, char *relpath, char *fullpath, char *destdir, struct s_dico *d)
{
	struct statvfs statvfsbuf;
	struct s_dico *footerdico;
	struct md5_ctx md5ctx;
	u64 curblocksize;
	char *origblock;
	u64 blkoffset;
	char textbuf[128];
	u8 md5sumcalc[16];
	u8 md5sumorig[16];
	int errors;
	u64 filesize;
	u64 filepos;
	int ret=0;
	long lres;
	int fd;
	
	// init
	errors=0;
	md5_init_ctx(&md5ctx);

	if (dico_get_u64(d, DISKITEMKEY_SIZE, &filesize)!=0)
	{	errprintf("Cannot read filesize from archive for file=[%s]\n", relpath);
		return -1;
	}
	
	fd=open(fullpath, O_RDWR|O_CREAT|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd<0)
	{	sysprintf("Cannot open %s for writing\n", relpath);
		return -2;
	}
	
	for (filepos=0; filepos<filesize; filepos+=curblocksize)
	{
		if (arblkqueue_dequeue_block_wait(&ai->queue, &origblock, &curblocksize, &blkoffset)!=0)
		{	errprintf("arblkqueue_dequeue_block for file(%s) failed\n", relpath);
			errors++;
			goto restore_obj_error;
		}
		
		if (blkoffset!=filepos)
		{	errprintf("file offset do not match for file(%s) failed\n", relpath);
			errors++;
			goto restore_obj_error;
		}
		
		md5_process_bytes(origblock, curblocksize, &md5ctx);
		if ((lres=write(fd, origblock, (long)curblocksize))!=curblocksize) // error
		{
			if ((lres>0) && (lres < curblocksize)) // probably "no space left"
			{
				if (fstatvfs(fd, &statvfsbuf)!=0)
				{	sysprintf("fstatvfs(fd=%d) failed\n", fd);
					ret=-5;
					goto restore_obj_error;
				}
				
				u64 freebytes = statvfsbuf.f_bfree * statvfsbuf.f_bsize;
				msgprintf(1, "Can't write all data for file [%s]: space on device is %s\n", relpath, 
					format_size(freebytes, textbuf, sizeof(textbuf), 'h'));
				ret=-5;
				goto restore_obj_error;
			}
			else // another error
			{
				sysprintf("cannot extract %s, curblocksize=%ld\n", relpath, (long)curblocksize);
				ret=-5;
				goto restore_obj_error;
			}
		}
		free(origblock);
	}
	
	// check the whole file md5 checksum
	md5_finish_ctx(&md5ctx, md5sumcalc);
	
	if ((footerdico=arblkqueue_dequeue_header_wait(&ai->queue, FDARCHIVER_MAGIC_FILEFOOTER))==NULL)
	{	errprintf("arblkqueue_dequeue_header_wait() failed: cannot read footer dico\n");
		errors++;
		goto restore_obj_error;
	}

	if (dico_get(footerdico, BLOCKFOOTITEMKEY_MD5SUM, md5sumorig, 16, NULL))
	{	errprintf("cannot get md5sum from file footer for file=[%s]\n", relpath);
		dico_show(footerdico);
		ret=-7;
		goto restore_obj_error;
	}
	
	if (memcmp(md5sumcalc, md5sumorig, 16)!=0)
	{	errprintf("cannot restore file %s, file is corrupt\n", relpath);
		errors++;
	}
	
restore_obj_error:
	dico_destroy(footerdico);
	close(fd);
	return ret;
}

int restore_object(struct s_archinfo *ai, int *errors, char *destdir, struct s_dico *dstd, struct s_dico *dxattr)
{
	char relpath[PATH_MAX];
	char fullpath[PATH_MAX];
	char parentdir[PATH_MAX];
	char xattrname[1024];
	char xattrvalue[1024];
	u16 xattrdatasize;
	u32 mode, uid, gid;
	u64 atime, mtime;
	struct utimbuf times;
	u32 objtype;
	int xattrdicsize;
	int res;
	int i;
	
	// init
	*errors=0;
	
	if (dico_get(dstd, DISKITEMKEY_PATH, relpath, sizeof(relpath), NULL)!=0)
		return -1;
	if (dico_get_u32(dstd, DISKITEMKEY_OBJTYPE, &objtype)!=0)
		return -2;
	concatenate_paths(fullpath, sizeof(fullpath), destdir, relpath);
	
	msgprintf(1, "-[%s] %s\n", get_objtype_name(objtype), relpath);
	
	// create parent directory first
	extract_dirpath(fullpath, parentdir, sizeof(parentdir));
	mkdir_recursive(parentdir);

	switch (objtype)
	{
		case OBJTYPE_DIR:
			dbgprintf(2, "objtype=OBJTYPE_DIR, path=[%s]\n", relpath);
			res=restore_obj_directory(ai, fullpath, destdir, dstd);
			break;
		case OBJTYPE_SYMLINK:
			dbgprintf(2, "objtype=OBJTYPE_SYMLINK, path=[%s]\n", relpath);
			res=restore_obj_symlink(ai, fullpath, destdir, dstd);
			break;
		case OBJTYPE_REGFILE:
			dbgprintf(2, "objtype=OBJTYPE_REGFILE, path=[%s]\n", relpath);
			res=restore_obj_regfile(ai, relpath, fullpath, destdir, dstd);
			if (res<0)
			{	errprintf("restore_obj_regfile(%s) failed with res=%d\n", relpath, res);
			}
			else if (res>0) // corruptions found
			{	(*errors)++;
				res=0; // continue anyway
			}
			break;
		case OBJTYPE_HARDLINK:
			dbgprintf(2, "objtype=OBJTYPE_HARDLINK, path=[%s]\n", relpath);
			res=restore_obj_hardlink(ai, fullpath, destdir, dstd);
			break;
		case OBJTYPE_CHARDEV:
			dbgprintf(2, "objtype=OBJTYPE_CHARDEV, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dstd);
			break;
		case OBJTYPE_BLOCKDEV:
			dbgprintf(2, "objtype=OBJTYPE_BLOCKDEV, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dstd);
			break;
		case OBJTYPE_FIFO:
			dbgprintf(2, "objtype=OBJTYPE_FIFO, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dstd);
			break;
		case OBJTYPE_SOCKET:
			dbgprintf(2, "objtype=OBJTYPE_SOCKET, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dstd);
			break;
		default:
			errprintf("Unknown objtype %d\n", objtype);
			res=-1;
			return -3;
	}
	
	// ---- restore standard attributes (permissions, owner, ...)
	if (res!=0) // value returned by restore_obj_xxx()
	{	errprintf("Restoring file=[%s], objtype=[%s] failed\n", fullpath, get_objtype_name(objtype));
		return -1;
	}
	if (dico_get_u32(dstd, DISKITEMKEY_MODE, &mode)!=0)
		return -4;
	if (dico_get_u32(dstd, DISKITEMKEY_UID, &uid)!=0)
		return -5;
	if (dico_get_u32(dstd, DISKITEMKEY_GID, &gid)!=0)
		return -6;
	if (dico_get_u64(dstd, DISKITEMKEY_ATIME, &atime)!=0)
		return -7;
	if (dico_get_u64(dstd, DISKITEMKEY_MTIME, &mtime)!=0)
		return -8;
	times.actime=(time_t)atime;
	times.modtime=(time_t)mtime;
	
	if (lchown(fullpath, (uid_t)uid, (gid_t)gid)!=0)
	{	sysprintf("Cannot lchown(%s) which is %s\n", fullpath, get_objtype_name(objtype));
		return -10;
	}
	
	if (objtype!=OBJTYPE_SYMLINK)
	{
		if (chmod(fullpath, (mode_t)mode)!=0)
			return -9;
	}

	// set the values for atime/mtime
	struct timeval tv[2];
	tv[0].tv_usec=0;
	tv[0].tv_sec=atime;
	tv[1].tv_usec=0;
	tv[1].tv_sec=mtime;
	if (lutimes(fullpath, tv)!=0)
		return -11;

	// ---- restore extended attributes
	xattrdicsize=dico_count(dxattr);

	for (i=0; i < xattrdicsize; i+=2)
	{
		if (dico_get_string(dxattr, (u64)(i+0), xattrname, sizeof(xattrname))!=0)
		{	errprintf("Cannot retrieve the name of an xattr for file %s\n", relpath);
			dico_show(dxattr);
			return -1;
		}
		memset(xattrvalue, 0, sizeof(xattrvalue));
		if (dico_get(dxattr, (u64)(i+1), xattrvalue, sizeof(xattrvalue), &xattrdatasize)!=0)
		{	errprintf("Cannot retrieve the value of an xattr for file %s\n", relpath);
			return -1;
		}
		
		res=lsetxattr(fullpath, xattrname, xattrvalue, xattrdatasize, 0);
		if (res!=0)
		{ sysprintf("lsetxattr(%s,%s,%s) failed\n", relpath, xattrname, xattrvalue); }
		else
		{ msgprintf(2, "            lsetxattr(%s, %s, %s)=%d\n", relpath, xattrname, xattrvalue, res); }
	}

	return 0;
}

int extract_read_objects(struct s_archinfo *ai, int *errors, char *destdir)
{
	char magic[FDARCHIVER_SIZEOF_MAGIC];
	struct s_dico *dicostd;
	struct s_dico *dicoxattr;
	int res;
	int ret=0;
	u32 objtype;
	int curerr;
	int foundobj;
	int type;
	
	// init
	*errors=0;
	
	do
	{	// skip the garbage (just ignore everything until the next FDARCHIVER_MAGIC_OBJT)
		// in case the archive is corrupt and contents has been added / removed in the archive
		do
		{
			if (arblkqueue_check_next_item_wait(&ai->queue, &type, magic)!=0)
			{	errprintf("arblkqueue_check_next_item() failed: cannot read object from archive\n");
				ret=-1;
				goto extract_read_objects_error;
			}
			
			if (memcmp(magic, FDARCHIVER_MAGIC_OBJT, FDARCHIVER_SIZEOF_MAGIC)==0)
			{
				foundobj=true; // normal case: header is what we expect
			}
			else // did not find expected header: skip garbage in archive
			{	msgprintf(1, "garbage found in archive, skipping it: type=%d, magic=[%s]\n", 
					type, (type==ARBLKITEM_TYPE_DICO)?(magic):"-block-");
				if (arblkqueue_destroy_first_item(&ai->queue)!=0)
				{	errprintf("arblkqueue_destroy_first_item() failed: cannot read object from archive\n");
					ret=-1;
					goto extract_read_objects_error;
				}
				foundobj=false;
			}
		} while (foundobj!=true);
		
		// read the object next header
		if ((dicostd=arblkqueue_dequeue_header_wait(&ai->queue, FDARCHIVER_MAGIC_OBJT))==NULL)
		{	errprintf("arblkqueue_dequeue_header_wait() failed: cannot read object dico\n");
			ret=-1;
			goto extract_read_objects_error;
		}

		res=dico_get_u32(dicostd, DISKITEMKEY_OBJTYPE, &objtype);
		if (objtype!=OBJTYPE_END)
		{
			// read item header with the extended attributes
			if ((dicoxattr=arblkqueue_dequeue_header_wait(&ai->queue, FDARCHIVER_MAGIC_XATTR))==NULL)
			{	errprintf("arblkqueue_dequeue_header_wait() failed: cannot read xattr dico\n");
				ret=-1;
				goto extract_read_objects_error;
			}

			res=restore_object(ai, &curerr, destdir, dicostd, dicoxattr);
			if (res!=0)
			{	errprintf("restore_object() failed with res=%d\n", res);
				ret=-1;
				goto extract_read_objects_error;
			}
			else if (curerr>0)
			{
				*errors+=curerr;	
			}
		}
		dico_destroy(dicoxattr);
		dico_destroy(dicostd);
	} while ((res==0) && (objtype!=OBJTYPE_END) && (get_terminate()==0));

extract_read_objects_error:
	return ret;
}

int mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	// create filesystem
	if ((strcmp(filesystem, "ext2")==0) || (strcmp(filesystem, "ext3")==0) || (strcmp(filesystem, "ext4")==0))
		return ext2_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "reiserfs")==0)
		return reiser_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "reiser4")==0)
		return reiser4_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "btrfs")==0)
		return btrfs_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "xfs")==0)
		return xfs_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "jfs")==0)
		return jfs_mkfs(d, partition, filesystem);
	else // unknown fs
		{ dbgprintf(1, "unknown filesystem: [%s]\n", filesystem); return -1; }

	return 0;
}

int read_mainhead(struct s_archinfo *ai)
{
	struct s_dico *d;
	char progvercreat[64];
	char fileformat[64];
	char label[64];

	if (!ai)
	{	dbgprintf(2, "ai is NULL\n");
		return -1;
	}

	if ((d=arblkqueue_dequeue_header_wait(&ai->queue, FDARCHIVER_MAGIC_MAIN))==NULL)
	{	dbgprintf(2, "arblkqueue_dequeue_header_wait() failed: cannot read main header\n");
		return -1;
	}
	
	if (dico_get_u32(d, MAINHEADKEY_ARCHIVEID, &ai->archid)!=0)
	{	dbgprintf(2, "cannot find archid in main-header\n");
		return -1;
	}

	if (dico_get(d, MAINHEADKEY_FILEFORMATVER, fileformat, sizeof(fileformat), NULL)!=0)
	{	dbgprintf(2, "cannot find fileformatver in main-header\n");
		return -1;
	}

	if (dico_get(d, MAINHEADKEY_PROGVERCREAT, progvercreat, sizeof(progvercreat), NULL)!=0)
	{	dbgprintf(2, "cannot find progvercreat in main-header\n");
		return -1;
	}

	if (dico_get(d, MAINHEADKEY_ARCHLABEL, label, sizeof(label), NULL)!=0)
	{	dbgprintf(2, "cannot find label in main-header\n");
		return -1;
	}

	if (strcmp(fileformat, FSARCHIVER_FILEFORMAT)!=0)
	{
		errprintf("This archive is based on a different file format: [%s]. Cannot continue.\n", fileformat);
		errprintf("It has been created with fsarchiver [%s], you should extrat the archive using that version.\n", progvercreat);
		errprintf("The current version of the program is [%s], and it's based on format [%s]\n", FSARCHIVER_VERSION, FSARCHIVER_FILEFORMAT);
		return -1;
	}

	return 0;
}

int extract_filesystem(struct s_archinfo *ai, char *partition)
{
	struct s_dico *dicofs;
	char mntbuf[PATH_MAX];
	char filesystem[128];
	u64 fsbytestotal;
	u64 fsbytesused;
	char text[128];
	char optbuf[128];
	char fsbuf[128];
	int readwrite;
	int errors;
	int ret=0;
	int res;
	
	// check the partition is not mounted
	res=mount_get_devinfo(partition, &readwrite, mntbuf, sizeof(mntbuf), optbuf, sizeof(optbuf), fsbuf, sizeof(fsbuf));
	if (res==0) 
	{	errprintf("partition [%s] is mounted. cannot restore an archive to a partition which is mounted.\n", partition);
		return -1;
	}
	
	// ---- read filesystem-header from archive
	if ((dicofs=arblkqueue_dequeue_header_wait(&ai->queue, FDARCHIVER_MAGIC_FSYS))==NULL)
	{	errprintf("arblkqueue_dequeue_header_wait() failed: cannot read file system dico\n");
		return -1;
	}
	
	dico_get_string(dicofs, FSYSHEADKEY_FILESYSTEM, filesystem, sizeof(filesystem));
	dico_get_u64(dicofs, FSYSHEADKEY_BYTESTOTAL, &fsbytestotal);
	dico_get_u64(dicofs, FSYSHEADKEY_BYTESUSED, &fsbytesused);
	snprintf(fsbuf, sizeof(fsbuf), "%s", filesystem);
	dbgprintf(2, "filesystem=[%s]\n", filesystem);
	dbgprintf(2, "fsbytestotal=[%s]\n", format_size(fsbytestotal, text, sizeof(text), 'h'));
	dbgprintf(2, "fsbytesused=[%s]\n", format_size(fsbytesused, text, sizeof(text), 'h'));
	
	// ---- make the filesystem
	if (g_options.destfs[0]) // a filesystem was specified in the cmd line
		snprintf(filesystem, sizeof(filesystem), "%s", g_options.destfs);
	
	// check if the destination filesystem is supported by the kernel
	if (check_filesystem_support(filesystem)!=0)
	{	errprintf("the filesystem %s does not seems to be supported by the current kernel (see /proc/filesystems)\n", filesystem);
		ret=-1;
		goto extract_filesystem_return;
	}
	
	if(mkfs(dicofs, partition, filesystem)!=0)
	{	errprintf("cannot format the filesystem %s on partition %s\n", filesystem, partition);
		ret=-1;
		goto extract_filesystem_return;
	}

	// ---- mount the new filesystem
	mkdir_recursive(mntbuf);
	generate_random_tmpdir(mntbuf, sizeof(mntbuf));
	mkdir_recursive(mntbuf);
	snprintf(fsbuf, sizeof(fsbuf), "%s", filesystem);
	res=mount_partition(partition, mntbuf, fsbuf, sizeof(fsbuf), 0);
	if (res!=0)
	{	errprintf("partition [%s] cannot be mounted on %s. cannot continue.\n", partition, mntbuf);
		ret=-1;
		goto extract_filesystem_return;
	}

	res=extract_read_objects(ai, &errors, mntbuf);
	if (res!=0)
	{	errprintf("extract_read_objects(%s) failed\n", mntbuf);
		ret=-1;
		goto extract_filesystem_umount;
	}
	else if (errors>0)
	{	errprintf("extract_read_objects(%s) worked with errors\n", mntbuf);
		ret=-1;
		goto extract_filesystem_umount;
	}

extract_filesystem_umount:
	if (unmount_partition(mntbuf)!=0)
	{	sysprintf("cannot umount %s\n", mntbuf);
		ret=-1;
	}
extract_filesystem_return:
	dico_destroy(dicofs);
	return ret;
}

int do_extract(char *archive, char **partition)
{
	pthread_t thread_decomp[FDARCHIVER_MAXJOBS];
	pthread_t thread_reader;
	struct s_archinfo ai;
	struct stat st;
	int ret=0;
	int i;

	// check the partitions are all block devices
	for (i=0; partition[i]; i++)
	{
		if ((stat(partition[i], &st)!=0) || (!S_ISBLK(st.st_mode)))
		{	errprintf("%s is not a valid block device\n", partition[i]);
			ret=-1;
			goto do_extract_destroy;
		}
	}
	
	// open archive file
	archive_init(&ai);
	if (archive_open(&ai, archive)!=0)
	{       errprintf ("archive_open(%s) failed\n", archive);
		ret=-1;
		goto do_extract_destroy;
	}
	
	// create decompression threads
	for (i=0; (i<g_options.compressjobs) && (i<FDARCHIVER_MAXJOBS); i++)
	{
		if (pthread_create(&thread_decomp[i], NULL, thread_decomp_fct, (void*)&ai) != 0)
		{       errprintf ("pthread_create(thread_decomp_fct) failed\n");
			ret=-1;
			goto do_extract_destroy;
		}
	}
	
	// create archive-reader thread
	if (pthread_create(&thread_reader, NULL, thread_reader_fct, (void*)&ai) != 0)
	{       errprintf ("pthread_create(thread_reader_fct) failed\n");
		ret=-1;
		goto do_extract_destroy;
	}
	
	if (read_mainhead(&ai)<0)
	{       errprintf ("archive_read_mainhead(%s) failed\n", archive);
		ret=-1;
		goto do_extract_close;
	}
	
	for (i=0; partition[i]; i++)
	{
		if (extract_filesystem(&ai, partition[i])!=0)
		{	errprintf("extract_filesystem(%s) failed\n", partition[i]);
			goto do_extract_close;
		}
	}
	
	ai.terminate=true; // ask threads to terminate
	
	for (i=0; (i<g_options.compressjobs) && (i<FDARCHIVER_MAXJOBS); i++)
	{
		if (pthread_join(thread_decomp[i], NULL) != 0)
		{       errprintf ("pthread_join(thread_decomp) failed\n");
			ret=-1;
			goto do_extract_destroy;
		}
	}
		
	if (pthread_join(thread_reader, NULL) != 0)
	{       errprintf ("pthread_join(thread_reader) failed\n");
		ret=-1;
		goto do_extract_destroy;
	}
	
do_extract_close:
	archive_close(&ai);
do_extract_destroy:
	archive_destroy(&ai);
	return ret;
}
