/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <linux/limits.h>
#include <sys/statvfs.h>

#include "fsarchiver.h"
#include "filereadbuf.h"
#include "common.h"

/* filereadbuf: this object implements buffered read from a file. It has been necessary to 
   fseek to the next header in the file without having to use fseek, so that we can read 
   from a pipe / stdin. For instance, when a file is corrupt, we want to move the file
   pointer to the beginning of the next header. But we cannot just read until we find
   the magic of the next header, and then fseek back to the beginning of the header,
   and call the read_header function, because fseek won't work on pipes. So we must ensure
   we only read each byte once and in the order. So we need such as buffered class.
*/

int filereadbuf_init(struct s_filereadbuf *frb, int fd)
{
	struct stat statfs;
	
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}
	
	memset(frb->ringbuf, 0, IMGREADBUF_BUFFER_SIZE);
	frb->filefd=fd;
	frb->buflen=0;
	frb->start=0;
	frb->eof=false;
	
	// get archive size
	if (fstat(fd, &statfs)!=0)
	{	errprintf("fstat() failed\n");
		return -1;
	}	
	else
	{
		frb->filesize=statfs.st_size;
	}
	
	return 0;
}

// say true if end-of-file is reached and buffer is empty
int filereadbuf_eof(struct s_filereadbuf *frb)
{
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}
	
	return ((frb->buflen==0) && (frb->eof==true));
}

// compare the beginning of the ring buffer with the data given in argv
int filereadbuf_compare(struct s_filereadbuf *frb, char *data, int len)
{
	int i;
	
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1; // error
	}
	
	if (frb->buflen < len)
		return false;

	for (i=0; i < len; i++)
		if (frb->ringbuf[((u64)(frb->start+i))%((u64)IMGREADBUF_BUFFER_SIZE)]!=data[i])
			return false; // different
	
	return true; // match
}

// fill the ringbuffer from file until buffer is full or eof
int filereadbuf_fill_buffer(struct s_filereadbuf *frb)
{
	u64 curpos;
	char c;

	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}
	
	curpos=(frb->start)+(frb->buflen);
	while ((curpos < frb->filesize) && (frb->buflen < IMGREADBUF_BUFFER_SIZE))
	{
		if (read(frb->filefd, &c, 1)!=1)
		{	sysprintf("read() failed\n");
			return -1;
		}
		frb->ringbuf[(curpos++)%((u64)IMGREADBUF_BUFFER_SIZE)]=c;
		frb->buflen++;
	}

	if (curpos>=frb->filesize)
		frb->eof=true;

	//fprintf(stderr, "filereadbuf_fill_buffer()=OK\n");
	return 0;
}

// fseek in the file buffer where we can find that pattern (an header magic)
int filereadbuf_goto_next_pattern(struct s_filereadbuf *frb, char *pattern, int patlen)
{
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}
	
	//fprintf(stderr, "\n\n--- filereadbuf_goto_next_pattern(%s)\n", pattern);
	while ((frb->buflen >= patlen) || (frb->eof!=true)) // can still read something
	{
		//fprintf(stderr, " {1} ");
		if ((frb->buflen < patlen) && (frb->eof!=true))
		{
			filereadbuf_fill_buffer(frb);
		}
		if (filereadbuf_compare(frb, pattern, patlen)==true)
			return 0;
		//fprintf(stderr, " {2} ");
		frb->start++;
		frb->buflen--;
		//fprintf(stderr, " {3} ");
	}
	//fprintf(stderr, "--- filereadbuf_goto_next_pattern(%s)\n\n\n", pattern);
	
	return 0;
}

// just read data from the file at the current pointer in file
int filereadbuf_read(struct s_filereadbuf *frb, void *buf, u64 size)
{
	char *buffer;
	u64 i;
	u64 pos;
	
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}
	buffer=buf;
	
	for (i=0; i < size; i++)
	{
		if (frb->buflen==0) // need more data
		{
			if (frb->eof==true)
				return -1;
			else
				filereadbuf_fill_buffer(frb);
		}
		pos=((u64)(frb->start))%((u64)IMGREADBUF_BUFFER_SIZE);
		buffer[i]=frb->ringbuf[(long)pos];
		frb->start++;
		frb->buflen--;
	}
	
	return 0;
}

// just read data from the file at the current pointer in file
int filereadbuf_read_nomove(struct s_filereadbuf *frb, void *buf, u64 size)
{
	char *buffer;
	u64 start;
	u64 buflen;
	u64 i;
	u64 pos;
	
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}
	buffer=buf;
	if (size > IMGREADBUF_BUFFER_SIZE)
	{	errprintf("size is too big: size=%ld and IMGREADBUF_BUFFER_SIZE=%d\n", (long)size, IMGREADBUF_BUFFER_SIZE);
		return -1;
	}
	
	filereadbuf_fill_buffer(frb);
	
	if (frb->buflen < size)
	{	errprintf("not enough data in the buffer: size=%ld and buflen=%ld\n", (long)size, (long)frb->buflen);
		return -1;
	}
	
	start=frb->start;
	buflen=frb->buflen;
	
	for (i=0; i < size; i++)
	{
		if (buflen==0) // need more data
		{
			if (frb->eof==true)
				return -1;
			else
				filereadbuf_fill_buffer(frb);
		}
		pos=(start)%((u64)IMGREADBUF_BUFFER_SIZE);
		buffer[i]=frb->ringbuf[(long)pos];
		start++;
		buflen--;
	}
	
	return 0;
}

// move the position of the pointer: sort of seek from the current position
int filereadbuf_increase_position(struct s_filereadbuf *frb, u64 size)
{
	frb->start+=size;
	frb->buflen-=size;

	return 0;
}

int filereadbuf_destroy(struct s_filereadbuf *frb)
{
	if (frb==NULL)
	{	errprintf("frb is NULL\n");
		return -1;
	}

	return 0;
}

/*int test_filereadbuf()
{
	struct s_filereadbuf frb;
	char buffer[1024];
	int fd=-1, res;

	fd=open("/etc/fstab", O_RDONLY);
	if (fd<0)
	{
		printf("open failed\n");
		return 0;
	}
	memset(buffer, 0, sizeof(buffer));
	res=filereadbuf_init(&frb, fd);
	res=filereadbuf_read(&frb, buffer, 10);
	printf("filereadbuf_read()=%d\n", res);
	printf("buffer=[%s]\n", buffer);
	filereadbuf_goto_next_pattern(&frb, "/dev/sda4", strlen("/dev/sda4"));
	res=filereadbuf_read(&frb, buffer, 30);
	printf("buffer-sda4=[%s]\n", buffer);

	filereadbuf_destroy(&frb);
	
	return 0;
}*/
