/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <endian.h>
#include <byteswap.h>
#include <signal.h>
#include <linux/limits.h>

#include "types.h"
#include "dico.h"
#include "fsarchiver.h"
#include "common.h"
#include "extract.h"
#include "create.h"

char *valid_magic[]={FDARCHIVER_MAGIC_MAIN, FDARCHIVER_MAGIC_FSYS, FDARCHIVER_MAGIC_OBJT, 
	FDARCHIVER_MAGIC_XATTR, FDARCHIVER_MAGIC_BLOCKHEADER, FDARCHIVER_MAGIC_FILEFOOTER, NULL};

struct s_options g_options;
FILE *g_logfile=NULL;
int g_stop=false;

int get_terminate() // return 1 if a SIGINT/SIGTERM has been received or a thread failed
{
	int mysigs[]={SIGINT, SIGTERM, -1};
	sigset_t mask_set;
	sigpending(&mask_set);
	int i;
	
	if (g_stop==true) // a thread exited with an error
		return 1;
	
	if (sigpending(&mask_set)==0)
	{	
		for (i=0; mysigs[i]!=-1; i++)
		{
			if (sigismember(&mask_set, mysigs[i]))
			{	dbgprintf(1, "get_terminate(): received signal %d\n", mysigs[i]);
				return 1;
			}
		}
	}
	
	return 0;
}

void usage(char *progname)
{
	msgprintf(0, "====> fsarchiver version %s (%s) - http://www.fsarchiver.org <====\n", FSARCHIVER_VERSION, FSARCHIVER_RELDATE);
	msgprintf(0, "Distributed under the GPL v2 license (GNU General Public License).\n");
	msgprintf(0, "<usage>\n");
	msgprintf(0, " * usage: %s [<options>] <command> <archive> <partition>\n", progname);
	msgprintf(0, " * eg: %s savefs /mnt/backup/mybackup.fsa /dev/sda1\n", progname);
	msgprintf(0, " * eg: %s restfs /mnt/backup/mybackup.fsa /dev/sda1\n", progname);
	msgprintf(0, "<commands>\n");
	msgprintf(0, " * savefs: save a filesystem to an archive file (backup a partition to a file)\n");
	msgprintf(0, " * restfs: restore a filesystem from an archive (overwrites the existing data)\n");
	msgprintf(0, "<options>\n");
	msgprintf(0, " -o: overwrite the archive if it already exists\n");
	msgprintf(0, " -v: verbose mode (can be used several times to increase the level of details)\n");
	msgprintf(0, " -d: debug mode (can be used several times to increase the level of details)\n");
	msgprintf(0, " -A: allow to save a filesystem which is mounted in read-write\n");
	msgprintf(0, " -F <filesystem>: extract the archive on a partition of that filesystem\n");
	msgprintf(0, " -z <level>: valid compression levels are: 2 (gzip, faster), 3 (bzip2, better)\n"); 
	msgprintf(0, " -j <count>: create more than one compression thread. useful on multi-core cpu\n"); 
	msgprintf(0, "Warnings:\n");
	msgprintf(0, " * this a development version, don't use for critical data.\n");
	msgprintf(0, " * the archive format will change in the future.\n");
}

int main(int argc, char **argv)
{
	sigset_t mask_set;
	char *command=NULL;
	char *archive=NULL;
	char *partition[32];
	char *progname;
	int ret=0;
	int c;

	// init
	progname=argv[0];
	memset(partition, 0, sizeof(partition));
	memset(&g_options, 0, sizeof(struct s_options));
	
	// set default options
	g_options.overwrite=0;
	g_options.allowsaverw=0;
	g_options.verboselevel=0;
	g_options.debuglevel=0;
	g_options.compressjobs=1;
	g_options.compresslevel=COMPRESS_GZIP;

	if (geteuid()!=0)
	{	errprintf("%s must be run as root. cannot continue.\n", progname);
		return 1;
	}
	
	while ((c = getopt(argc, argv, "oAF:vdz:j:")) != -1)
	{
		switch (c)
		{
			case 'o': // overwrite existing archives
				g_options.overwrite=1;
				break;
			case 'A': // allows to backup read/write mounted partition
				g_options.allowsaverw=1;
				break;
			case 'F': // use another filesystem on the dest partition
				snprintf(g_options.destfs, sizeof(g_options.destfs), "%s", optarg);
				break;
			case 'v': // verbose mode
				g_options.verboselevel++;
				break;
			case 'd': // debug mode
				g_options.debuglevel++;
				break;
			case 'j': // compression jobs
				g_options.compressjobs=atoi(optarg);
				if (g_options.compressjobs<1 || g_options.compressjobs>FDARCHIVER_MAXJOBS)
				{
					errprintf("[%s] is not a valid job number. Must be between 1 and %d\n", 
						optarg, FDARCHIVER_MAXJOBS);
					usage(progname);
					return 1;
				}
				break;
			case 'z': // compression level
				switch (atoi(optarg))
				{
					case 2:
						g_options.compresslevel=COMPRESS_GZIP;
						break;
					case 3:
						g_options.compresslevel=COMPRESS_BZIP2;
						break;
					default:
						errprintf("[%s] is not a valid compression level\n", optarg);
						usage(progname);
						return 1;
				};
				break;
			default:
				usage(progname);
				return 1;
		}
	}
	
	argc -= optind;
	argv += optind;
	
	if (argc < 3)
	{	errprintf("there are not enough arguments on the command line\n");
		usage(progname);
		return 1;
	}
	
	// open debug logfile if requested
	if (g_options.debuglevel>0)
	{	mkdir_recursive("/var/log");
		msgprintf(1, "g_options.debuglevel=%d\n", g_options.debuglevel);
		g_logfile=fopen("/var/log/fsarchiver.log", "w+");
		if (g_logfile)
		{ msgprintf(1, "Creating logfile in /var/log/fsarchiver.log\n");}
		else
		{ sysprintf("Cannot create logfile in /var/log/fsarchiver.log\n");}
	}

	command=*argv++, argc--;
	archive=*argv++, argc--;
	
	// TODO: allow more than one partition per archive
	//for (i=0; (i < argc) && (argv[i]); i++)
	partition[0]=argv[0];

	// install signal handlers
	sigemptyset(&mask_set);
	sigaddset(&mask_set, SIGINT);
	sigaddset(&mask_set, SIGTERM);
	sigprocmask(SIG_SETMASK, &mask_set, NULL);
	
	if (strcmp(command, "savefs")==0)
		ret=do_create(archive, partition);
	else if (strcmp(command, "restfs")==0)
		ret=do_extract(archive, partition);
	else // command not found
	{	errprintf("[%s] is not a valid command.\n", command);
		usage(progname);
		ret=1;
	}
	
	// close the debug logfile if it has been open
	if ((g_options.debuglevel>0) && (g_logfile!=NULL))
		fclose(g_logfile);

	return ret;
}
