/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>

#include "fsarchiver.h"
#include "archive.h"
#include "common.h"
#include "comp_gzip.h"
#include "comp_bzip2.h"

int compress_block(void *bufdat, u64 bufsize, void *origblock, u64 curblocksize, u64 offset, u64 *newsize)
{
	u8 compblock[DATA_MAX_BLOCK_SIZE];
	struct s_blockheader header;
	u32 finalblockcsum;
	u8 *finaldata;
	u8 *curpos;
	u64 finalblksize;
	u64 compsize;
	u32 algoused;
	int res;
	
	// init
	*newsize=0;
	memset(&header, 0, sizeof(header));
	memcpy(header.magic, "blkh", 4);
	header.blocksize=cpu_to_le64(curblocksize);
	header.blockoffset=cpu_to_le64(offset);
	
	// -------------- compress the block
	switch (g_options.compresslevel)
	{
		case COMPRESS_GZIP:
			res=compress_block_gzip(curblocksize, &compsize, origblock, curblocksize, compblock, sizeof(compblock));
			algoused=COMPRESS_GZIP;
			break;
		case COMPRESS_BZIP2:
			res=compress_block_bzip2(curblocksize, &compsize, origblock, curblocksize, compblock, sizeof(compblock));
			algoused=COMPRESS_BZIP2;
			break;
		default:
			compsize=sizeof(compblock);
			algoused=COMPRESS_NULL;
			dbgprintf(1, "invalid compression level: %d\n", (int)g_options.compresslevel);
			return -1;
	}

	// check compression status and efficiency
	if ((res==0) && (compsize<curblocksize)) // compression worked and was good
	{	finaldata=compblock;
		finalblksize=compsize;
	}
	else // compressed version is bigger: write the original block
	{	algoused=COMPRESS_NONE;
		finaldata=origblock;
		finalblksize=curblocksize;
	}
	
	// calculate the final checksum
	finalblockcsum=fletcher32(finaldata, finalblksize);
	header.finalcsum=cpu_to_le32(finalblockcsum);
	header.finalsize=cpu_to_le64(finalblksize);
	header.compalgo=cpu_to_le32(algoused);
	
	// write header and newdata to buffer
	*newsize=sizeof(header)+finalblksize;
	memset(bufdat, 0, bufsize);
	if (*newsize > bufsize)
	{
		errprintf("buffer too small: bufsize=%ld, finalblksize=%ld\n", (long)bufsize, (long)finalblksize);
		return -1;
	}
	
	// prepare buffer with header+data
	curpos=bufdat;
	curpos=mempcpy(curpos, &header, sizeof(header));
	curpos=mempcpy(curpos, finaldata, finalblksize);
	
	return 0;
}

int decompress_block(void *bufdat, u64 bufsize, void *origblock, u64 curblocksize, u64 *newsize)
{
	struct s_blockheader header;
	u32 finalblockcsumorig;
	u32 finalblockcsumcalc;
	u64 checkorigsize;
	u32 compalgo;
	u64 finalsize;
	u64 blockoffset;
	u8 *compdata;
	int errors;
	int res;
	
	// ---- init
	*newsize=0;

	// ---- read block header
	if (curblocksize < sizeof(header))
	{	errprintf("not enough data received\n");
		return -1;
	}
	
	memcpy(&header, origblock, sizeof(header));
	compdata=(u8*)origblock+sizeof(header);
	if (memcmp(header.magic, "blkh", 4)!=0)
	{	errprintf("invalid magic in header, expected [blkh]\n");
		return -1;
	}
	
	*newsize=le64_to_cpu(header.blocksize);
	blockoffset=le64_to_cpu(header.blockoffset);
	compalgo=le32_to_cpu(header.compalgo);
	finalsize=le64_to_cpu(header.finalsize);
	finalblockcsumorig=le32_to_cpu(header.finalcsum);
	
	// ---- read the compressed block from buffer
	if (curblocksize != sizeof(header)+(long)finalsize)
	{	errprintf("amount of data received is not what was expected\n");
		return -1;
	}
	if (bufsize < finalsize)
	{	errprintf("buffer too small for block data\n");
		return -1;
	}
	
	// check the block checksum
	finalblockcsumcalc=fletcher32(compdata, finalsize);
	if (finalblockcsumcalc!=finalblockcsumorig)
	{
		errprintf("block is corrupt at blockoffset=%ld, blksize=%ld\n", (long)blockoffset, (long)*newsize);
		memset(bufdat, 0, *newsize);
		errors++;
	}
	else // data not corrupted, decompresses the block
	{
		switch (compalgo)
		{
			case COMPRESS_NONE:
				memcpy(bufdat, compdata, *newsize);
				break;
			case COMPRESS_GZIP:
				if ((res=uncompress_block_gzip(finalsize, &checkorigsize, bufdat, bufsize, compdata))!=0)
				{	errprintf("uncompress_block_gzip()=%d failed: finalsize=%ld and checkorigsize=%ld\n", res, (long)finalsize, (long)checkorigsize);
					memset(origblock, 0, *newsize);
				}
				break;
			case COMPRESS_BZIP2:
				if ((res=uncompress_block_bzip2(finalsize, &checkorigsize, bufdat, bufsize, compdata))!=0)
				{	errprintf("uncompress_block_bzip2()=%d failed: finalsize=%ld and checkorigsize=%ld\n", res, (long)finalsize, (long)checkorigsize);
					memset(origblock, 0, *newsize);
				}
				break;
			default:
				errprintf("unsupported compression algorithm: %ld\n", (long)compalgo);
				return -1;
		}
	}

	return 0;
}

void *thread_comp_fct(void *args)
{
	char *origbuffer=NULL;
	char *compbuffer=NULL;
	struct s_archinfo *ai;
	u64 newsize;
	s64 blknum;
	u64 offset;
	u64 size;
	
	ai=(struct s_archinfo *)args;
	if (ai==NULL)
	{	errprintf("ai is NULL\n");
		goto thread_comp_fct_error;
	}
	
	while (ai->terminate==false || arblkqueue_count_status(&ai->queue, ARBLKITEM_STATUS_TODO)>0)
	{
		if ((blknum=arblkqueue_get_first_block_todo(&ai->queue, &origbuffer, &size, &offset))<0)
		{	errprintf("arblkqueue_dequeue_first_todo() failed\n");
			goto thread_comp_fct_error;
		}
		if (blknum>0) // block found
		{
			if (!(compbuffer=malloc(DATA_MAX_BLOCK_SIZE+1024)))
			{	errprintf("malloc() failed: cannot allocate memory for compressed block\n");
				goto thread_comp_fct_error;
			}
			if (compress_block(compbuffer, DATA_MAX_BLOCK_SIZE+1024, origbuffer, size, offset, &newsize)!=0)
			{	errprintf("compress_block() failed\n");
				goto thread_comp_fct_error;
			}

			if (arblkqueue_replace_block(&ai->queue, blknum, compbuffer, newsize, ARBLKITEM_STATUS_DONE)!=0)
			{	errprintf("arblkqueue_replace_block() failed\n");
				goto thread_comp_fct_error;
			}
			free(origbuffer);
		}
		else // blknum==1 means block not found
		{
			usleep(QUEUE_WAIT);
		}
	}

thread_comp_fct_error:
	g_stop=true;
	return NULL;
}

void *thread_decomp_fct(void *args)
{
	char *origbuffer;
	char *decompbuffer;
	struct s_archinfo *ai;
	u64 newsize;
	u64 offset;
	s64 blknum;
	u64 size;
	
	ai=(struct s_archinfo *)args;
	if (ai==NULL)
	{	errprintf("ai is NULL\n");
		goto thread_comp_fct_error;
	}
	
	while (ai->terminate==false || arblkqueue_count_status(&ai->queue, ARBLKITEM_STATUS_TODO)>0)
	{
		if ((blknum=arblkqueue_get_first_block_todo(&ai->queue, &origbuffer, &size, &offset))<0)
		{	errprintf("arblkqueue_dequeue_first_todo() failed\n");
			goto thread_comp_fct_error;
		}
		if (blknum>0) // block found
		{
			if (!(decompbuffer=malloc(DATA_MAX_BLOCK_SIZE+1024)))
			{	errprintf("malloc() failed: cannot allocate memory for compressed block\n");
				goto thread_comp_fct_error;
			}
			if (decompress_block(decompbuffer, DATA_MAX_BLOCK_SIZE+1024, origbuffer, size, &newsize)!=0)
			{	errprintf("decompress_block() failed\n");
				goto thread_comp_fct_error;
			}
			
			if (arblkqueue_replace_block(&ai->queue, blknum, decompbuffer, newsize, ARBLKITEM_STATUS_DONE)!=0)
			{	errprintf("arblkqueue_replace_block() failed\n");
				goto thread_comp_fct_error;
			}
			free(origbuffer);
		}
		else // block not found
		{
			usleep(QUEUE_WAIT);
		}
	}

thread_comp_fct_error:
	g_stop=true;
	return NULL;
}
