/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <linux/limits.h>
 #include <sys/statvfs.h>

#include "fsarchiver.h"
#include "archive.h"
#include "common.h"
#include "dico.h"

int imgdisk_write_data(struct s_archinfo *ai, void *data, u64 size)
{
	struct statvfs statvfsbuf;
	char textbuf[128];
	long lres;
	
	if (!ai)
	{	dbgprintf(2, "ai is NULL\n");
		return -1;
	}
	
	if ((lres=write(ai->archfd, (char*)data, (long)size))!=(long)size)
	{
		if ((lres>0) && (lres < size)) // probably "no space left"
		{
			if (fstatvfs(ai->archfd, &statvfsbuf)!=0)
			{	sysprintf("fstatvfs(fd=%d) failed\n", ai->archfd);
				return -1;
			}
			
			u64 freebytes = statvfsbuf.f_bfree * statvfsbuf.f_bsize;
			msgprintf(1, "Can't write archive file: space on device is %s\n", 
				format_size(freebytes, textbuf, sizeof(textbuf), 'h'));
			//msgprintf(2, "lres=%ld and curblocksize=%ld\n", (long)lres, (long)curblocksize);
			return -1;
		}
		else // another error
		{
			sysprintf("write() failed\n");
			return -1;
		}
	}
	
	return 0;
}

int imgdisk_read_data(struct s_archinfo *ai, void *data, u64 size)
{
	if (!ai)
	{	dbgprintf(2, "ai is NULL\n");
		return -1;
	}

	if (read(ai->archfd, (char*)data, (long)size)!=(long)size)
	{	sysprintf("read(size=%ld) failed\n", (long)size);
		return -1;
	}

	return 0;
}

int imgdisk_write_dico(struct s_archinfo *ai, struct s_dico *d)
{
	struct s_item *item;
	u32 headerlen;
	u32 checksum;
	u8 *buffer;
	u8 *bufpos;
	u16 temp16;
	u32 temp32;
	u64 temp64;
	u16 count;
	
	if (d==NULL)
	{	dbgprintf(1, "invalid dictionnary\n");
		return -1;
	}
	
	// 1. how many valid items there are
	count=dico_count(d);

	// 2. calculate len of header
	headerlen=sizeof(u16); // count
	for (item=d->head; item!=NULL; item=item->next)
	{
		if ((item->data) && (item->size>0))
		{
			headerlen+=sizeof(u64); // key
			headerlen+=sizeof(u16); // data size
			headerlen+=item->size; // data
		}
	}

	// 3. allocate memory for header
	bufpos=buffer=malloc(headerlen);
	if (!buffer)
	{	errprintf("cannot allocate memory for buffer");
		return -1;
	}
	
	// 4. write items count in buffer
	temp16=cpu_to_le16(count);
	bufpos=mempcpy(bufpos, &temp16, sizeof(temp16));
	
	// 5. write all items in buffer
	for (item=d->head; item!=NULL; item=item->next)
	{
		if ((item->data) && (item->size>0))
		{
			// 1. write key to buffer
			temp64=cpu_to_le64((u64)item->key);
			bufpos=mempcpy(bufpos, &temp64, sizeof(temp64));
			
			// 2. write sizeof(data) to buffer
			temp16=cpu_to_le16(item->size);
			bufpos=mempcpy(bufpos, &temp16, sizeof(temp16));
			
			// 3. write data to buffer
			bufpos=mempcpy(bufpos, item->data, item->size);
		}
	}

	// 6. write header-len, header-data, header-checksum
	temp32=cpu_to_le32(headerlen);
	if (imgdisk_write_data(ai, &temp32, sizeof(temp32))!=0)
	{	free(buffer);
		return -1;
	}
	
	if (imgdisk_write_data(ai, buffer, headerlen)!=0)
	{	free(buffer);
		return -1;
	}
	
	checksum=fletcher32(buffer, headerlen);
	temp32=cpu_to_le32(checksum);
	if (imgdisk_write_data(ai, &temp32, sizeof(temp32))!=0)
	{	free(buffer);
		return -1;
	}
	
	free(buffer);
	return 0;
}

int imgdisk_write_header(struct s_archinfo *ai, struct s_dico *d, char *magic)
{
	u32 temp32;
	
	if (!ai)
	{	dbgprintf(2, "ai is NULL\n");
		return -1;
	}
	
	if (d==NULL)
	{	dbgprintf(2, "invalid dictionnary\n");
		return -1;
	}
	
	// A. write dico magic string
	if (imgdisk_write_data(ai, magic, FDARCHIVER_SIZEOF_MAGIC)!=0)
		return -2;
	
	// B. write archive id
	temp32=cpu_to_le32(ai->archid);
	if (imgdisk_write_data(ai, &temp32, sizeof(temp32))!=0)
		return -3;
	
	// C. write the dico of the header
	if (imgdisk_write_dico(ai, d) != 0)
		return -4;
	
	return 0;
}

int imgdisk_write_block(struct s_archinfo *ai, char *blkdata, u64 blksize)
{
	struct s_dico *blkdico; // header written in file
	struct s_blockheader header; // internal header (used between threads)
	char *realdata;
	int res;
	
	// init
	memset(&header, 0, sizeof(header));
	if ((blkdico=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return -1;
	}
	
	// split raw block into header + realdata
	memcpy(&header, blkdata, sizeof(header));
	realdata=blkdata+sizeof(header);
	
	// prepare header
	dico_add_u64(blkdico, BLOCKHEADITEMKEY_BLOCKSIZE, header.blocksize);
	dico_add_u64(blkdico, BLOCKHEADITEMKEY_BLOCKOFFSET, header.blockoffset);
	dico_add_u64(blkdico, BLOCKHEADITEMKEY_FINALSIZE, header.finalsize);
	dico_add_u32(blkdico, BLOCKHEADITEMKEY_FINALCSUM, header.finalcsum);
	dico_add_u32(blkdico, BLOCKHEADITEMKEY_COMPRESSALGO, header.compalgo);
	
	if (sizeof(header)+header.finalsize != blksize)
	{	errprintf("received invalid block data/header\n");
		return -1;
	}
	
	// write block header
	res=imgdisk_write_header(ai, blkdico, FDARCHIVER_MAGIC_BLOCKHEADER);
	dico_destroy(blkdico);
	if (res!=0)
	{	errprintf("cannot write FDARCHIVER_MAGIC_BLOCKHEADER block-header\n");
		return -1;
	}
	
	// D. write block data
	if (imgdisk_write_data(ai, realdata, header.finalsize)!=0)
	{	errprintf("cannot write data block: imgdisk_write_data() failed\n");
		return -1;
	}
	
	return 0;
}

void *thread_writer_fct(void *args)
{
	char magic[FDARCHIVER_SIZEOF_MAGIC];
	char *blkdata;
	struct s_archinfo *ai;
	struct s_dico *d;
	u64 offset;
	s64 blknum;
	int type;
	u64 size;
	
	ai=(struct s_archinfo *)args;
	if (ai==NULL)
	{	errprintf("ai is NULL\n");
		goto thread_writer_fct_error;
	}
	
	while (ai->terminate==false || arblkqueue_count(&ai->queue)>0)
	{
		if ((blknum=arblkqueue_dequeue_first_if_done(&ai->queue, &type, &blkdata, &size, &offset, magic, &d))<0)
		{	errprintf("arblkqueue_dequeue_first_if_done() failed\n");
			goto thread_writer_fct_error;
		}

		if (blknum>0) // block found
		{
			if (type==ARBLKITEM_TYPE_BLOCK)
			{
				if (imgdisk_write_block(ai, blkdata, size)!=0)
				{	errprintf("imgdisk_write_block() failed\n");
					goto thread_writer_fct_error;
				}
				free(blkdata);
			}
			else if (type==ARBLKITEM_TYPE_DICO)
			{
				if (imgdisk_write_header(ai, d, magic)!=0)
				{	errprintf("imgdisk_write_header() failed\n", (long)size);
					goto thread_writer_fct_error;
				}
				dico_destroy(d);
			}
			else
			{
				fprintf(stderr, "WRIT: error-001\n");
			}
		}
		else // block not found
		{
			usleep(QUEUE_WAIT);
		}
	}
	
	goto thread_writer_fct_success;

thread_writer_fct_error:
	errprintf("thread_writer_fct failed\n");
	g_stop=true;
thread_writer_fct_success:
	return NULL;
}

// ================================================================================================================
// ================================================================================================================
// ================================================================================================================

int imgdisk_read_dico(struct s_archinfo *ai, struct s_dico *d)
{
	char *data;
	u16 size;
	u32 headerlen;
	u32 origsum;
	u32 newsum;
	u8 *buffer;
	u8 *bufpos;
	u16 temp16;
	u32 temp32;
	u64 temp64;
	u16 count;
	u64 key;
	int i;
	
	if (d==NULL)
	{	errprintf("invalid dictionnary\n");
		return -1;
	}
	
	// B. header-len, header-data, header-checksum
	if (imgdisk_read_data(ai, &temp32, sizeof(temp32))!=0)
		return -1;
	headerlen=le32_to_cpu(temp32);
	if (headerlen > FDARCHIVER_MAX_HEADER_SIZE)
	{	errprintf("invalid headerlen: %ld\n", (long)headerlen);
		return -2;
	}
	
	bufpos=buffer=malloc(headerlen);
	if (!buffer)
	{	errprintf("cannot allocate memory for header\n");
		return -2;
	}
	
	if (imgdisk_read_data(ai, buffer, headerlen)!=0)
	{	errprintf("cannot read header data\n");
		free(buffer);
		return -3;
	}
	
	if (imgdisk_read_data(ai, &temp32, sizeof(temp32))!=0)
	{	errprintf("cannot read header checksum\n");
		free(buffer);
		return -4;
	}
	origsum=le32_to_cpu(temp32);
	
	// check header-data integrity using checksum	
	newsum=fletcher32(buffer, headerlen);
	if (newsum!=origsum)
	{	errprintf("bad checksum for header\n");
		free(buffer);
		return -5;
	}
	
	// read count from buffer
	memcpy(&temp16, bufpos, sizeof(temp16));
	bufpos+=sizeof(temp16);
	count=le16_to_cpu(temp16);
	
	// C. read items
	for (i=0; i < count; i++)
	{
		// 1. read key from buffer
		memcpy(&temp64, bufpos, sizeof(temp64));
		bufpos+=sizeof(temp64);
		key=(u64)le64_to_cpu(temp64);
		
		// 2. read sizeof(data)
		memcpy(&temp16, bufpos, sizeof(temp16));
		bufpos+=sizeof(temp16);
		size=le16_to_cpu(temp16);
		if (size>DICO_MAX_DATASIZE) // invalid length
			return -8;

		if ((data=malloc(size))==NULL) // out of memory
			return -9;
		
		// 3. read data
		memcpy(data, bufpos, size);
		bufpos+=size;

		if (dico_add(d, key, data, size)!=0)
			return -10;
	}

	free(buffer);
	return 0;
}

struct s_dico *imgdisk_read_header(struct s_archinfo *ai, char *magic)
{
	struct s_dico *d;
	s64 curpos;
	u32 temp32;
	u32 archid;
	
	// init
	memset(magic, 0, FDARCHIVER_SIZEOF_MAGIC);

	if (!ai)
	{	errprintf("ai is NULL\n");
		return NULL;
	}
	
	if ((d=dico_alloc())==NULL)
	{	errprintf ("dico_alloc() failed\n");
		return NULL;
	}
	
	// A. search for next read header marker and magic (it may be further if corruption in archive)
	if ((curpos=lseek64(ai->archfd, 0, SEEK_CUR))<0)
	{	errprintf("lseek64() failed to get the current position in archive\n");
		return NULL;
	}
	if (imgdisk_read_data(ai, magic, FDARCHIVER_SIZEOF_MAGIC)!=0)
	{	errprintf("cannot read header magic\n");
		return NULL;
	}
	
	while (is_magic_valid(magic)!=true)
	{
		if (lseek64(ai->archfd, curpos++, SEEK_SET)<0)
		{	errprintf("lseek64() failed\n");
			return NULL;
		}
		if (imgdisk_read_data(ai, magic, FDARCHIVER_SIZEOF_MAGIC)!=0)
		{	errprintf("cannot read header magic\n");
			return NULL;
		}
	}
	
	// C. read the archive id
	if (imgdisk_read_data(ai, &temp32, sizeof(temp32))!=0)
	{	errprintf("cannot read archive-id in header\n");
		return NULL;
	}
	archid=le32_to_cpu(temp32);
	if (ai->archid) // only check archive-id if it's known (when main header has been read)
	{
		if (archid!=ai->archid)
		{	errprintf("archive-id in header does not match: archid=[%.8x], expected=[%.8x]\n",
				archid, ai->archid);
			return NULL;
		}
	}
	
	// D. read the dico of the header
	if (imgdisk_read_dico(ai, d) != 0)
	{	errprintf("imgdisk_read_dico() failed\n");
		return NULL;
	}
	//fprintf(stderr, "----------> found [%s]\n", magic);
	
	return d;
}

int read_and_copy_block(struct s_archinfo *ai, struct s_dico *blkdico, int *sumok)
{
	u8 *buffer;
	u8 *realdata;
	u64 totalsize; // internal head + data
	u64 curblocksize; // data size
	u64 blockoffset; // offset of the block in the file
	u32 compalgo; // compression algo used
	u64 finalsize; // compressed  block size
	u32 finalblockcsumorig;
	u32 finalblockcsumcalc;
	struct s_blockheader header; // internal header (used between threads)
	
	// init
	*sumok=-1;
	memset(&header, 0, sizeof(header));
	
	if (dico_get_u64(blkdico, BLOCKHEADITEMKEY_BLOCKSIZE, &curblocksize)!=0 || curblocksize>DATA_MAX_BLOCK_SIZE)
	{	dbgprintf(2, "cannot get blocksize from block-header\n");
		return -1;
	}
	
	if (dico_get_u64(blkdico, BLOCKHEADITEMKEY_BLOCKOFFSET, &blockoffset)!=0)
	{	dbgprintf(2, "cannot get blockoffset from block-header\n");
		return -1;
	}
	
	if (dico_get_u32(blkdico, BLOCKHEADITEMKEY_COMPRESSALGO, &compalgo)!=0)
	{	dbgprintf(2, "cannot get compr from block-header\n");
		return -1;
	}
	
	if (dico_get_u64(blkdico, BLOCKHEADITEMKEY_FINALSIZE, &finalsize)!=0)
	{	dbgprintf(2, "cannot get BLOCKHEADITEMKEY_FINALSIZE from block-header\n");
		return -1;
	}
	
	if (dico_get_u32(blkdico, BLOCKHEADITEMKEY_FINALCSUM, &finalblockcsumorig)!=0)
	{	dbgprintf(2, "cannot get BLOCKHEADITEMKEY_FINALCSUM from block-header\n");
		return -1;
	}
	
	// ---- allocate memory
	totalsize=sizeof(header)+finalsize;
	buffer=malloc(totalsize);
	if (!buffer)
	{	errprintf("cannot allocate block\n");
		return -1;
	}
	
	// prepare internal header
	memcpy(header.magic, "blkh", 4);
	header.blocksize=cpu_to_le64(curblocksize);
	header.blockoffset=cpu_to_le64(blockoffset);
	header.finalcsum=cpu_to_le32(finalblockcsumorig);
	header.finalsize=cpu_to_le64(finalsize);
	header.compalgo=cpu_to_le32(compalgo);
	realdata=mempcpy(buffer, &header, sizeof(header));
	
	if (read(ai->archfd, realdata, (long)finalsize)!=(long)finalsize)
	{	sysprintf("cannot read block (finalsize=%ld) failed\n", (long)finalsize);
		free(buffer);
		return -1;
	}
	
	// ---- checksum
	finalblockcsumcalc=fletcher32(realdata, finalsize);
	if (finalblockcsumcalc!=finalblockcsumorig)
	{
		errprintf("block is corrupt at offset=%ld, blksize=%ld\n", (long)blockoffset, (long)curblocksize);
		memset(buffer, 0, finalsize);
		free(buffer);
		buffer=malloc(curblocksize);
		if (!buffer)
		{	errprintf("cannot allocate block\n");
			return -1;
		}
		memset(buffer, 0, curblocksize);
		*sumok=false;
		if (arblkqueue_add_block_wait(&ai->queue, (char*)buffer, curblocksize, blockoffset, ARBLKITEM_STATUS_DONE)!=0)
		{	errprintf("arblkqueue_add_block() failed\n");
			return -1;
		}
		// go to the beginning of the corrupted contents so that the next header is searched here
		if (lseek64(ai->archfd, -(long)finalsize, SEEK_CUR)<0)
		{	errprintf("lseek64() failed\n");
			//return -1; --> do not stop, we can just continue to the next file buy may loose some data
		}
	}
	else // no corruption detected
	{	*sumok=true;
		if (arblkqueue_add_block_wait(&ai->queue, (char*)buffer, totalsize, blockoffset, ARBLKITEM_STATUS_TODO)!=0)
		{	errprintf("arblkqueue_add_block() failed\n");
			return -1;
		}
	}
	
	return 0;
}

void *thread_reader_fct(void *args)
{
	char magic[FDARCHIVER_SIZEOF_MAGIC];
	struct s_archinfo *ai;
	struct stat statfs;
	struct s_dico *d;
	u64 archsize;
	s64 curpos=0;
	int sumok;
	u64 errors;
	
	// init
	errors=0;
	ai=(struct s_archinfo *)args;
	if (ai==NULL)
	{	errprintf("ai is NULL\n");
		goto thread_reader_fct_error;
	}
	
	// get archive size
	if (fstat(ai->archfd, &statfs)!=0)
	{	errprintf("fstat() failed\n");
		goto thread_reader_fct_error;
	}
	archsize=statfs.st_size;
	
	// ---- read main archive header
	if ((d=imgdisk_read_header(ai, magic))==NULL)
	{	errprintf ("imgdisk_read_header() failed to read the archive header\n");
		goto thread_reader_fct_error;
	}
	if (dico_get_u32(d, MAINHEADKEY_ARCHIVEID, &ai->archid)!=0)
	{	dbgprintf(2, "cannot get archive-id from main header\n");
		goto thread_reader_fct_error;
	}
	
	if (arblkqueue_add_dico(&ai->queue, d, magic)!=0)
	{	errprintf("arblkqueue_add_dico() failed to add the archive header\n");
		goto thread_reader_fct_error;
	}
	
	// read all other data from file (filesys-header, normal objects headers, ...)
	while ((ai->terminate==false) && ((curpos=lseek64(ai->archfd, 0, SEEK_CUR))>=0) && (curpos < archsize))
	{
		if ((d=imgdisk_read_header(ai, magic))==NULL)
		{	errprintf ("imgdisk_read_header() failed\n");
			goto thread_reader_fct_error;
		}
		
		// if next thing in archive is a block, process block, else add header to the queue
		if (strncmp(magic, FDARCHIVER_MAGIC_BLOCKHEADER, FDARCHIVER_SIZEOF_MAGIC)==0)
		{
			//int read_and_copy_block(struct s_archinfo *ai, struct s_dico *blkdico, char *relpath, int *sumok)
			if (read_and_copy_block(ai, d, &sumok)!=0)
			{	errprintf ("read_and_copy_blocks() failed\n");
				goto thread_reader_fct_error;
			}
			if (sumok==false) errors++;
			dico_destroy(d);
		}
		else // it's a dico/header
		{
			if (arblkqueue_add_dico(&ai->queue, d, magic)!=0)
			{	errprintf("arblkqueue_add_dico() failed\n");
				goto thread_reader_fct_error;
			}
		}
	}
	
	goto thread_reader_fct_success;

thread_reader_fct_error:
	errprintf("thread_reader_fct failed\n");
	g_stop=true;
thread_reader_fct_success:
	return NULL;
}
